#!/usr/bin/python3

# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from datetime import datetime

import pytest

import pytest_custom_markers

pytest.register_assert_rewrite("generic")
import generic

pytestmark = pytest_custom_markers.have_json_c
requests = pytest.importorskip("requests")


# JSON helper functions
def fetch_zones_json(statsip, statsport):
    r = requests.get(
        "http://{}:{}/json/v1/zones".format(statsip, statsport), timeout=600
    )
    assert r.status_code == 200

    data = r.json()
    return data["views"]["_default"]["zones"]


def fetch_traffic_json(statsip, statsport):
    r = requests.get(
        "http://{}:{}/json/v1/traffic".format(statsip, statsport), timeout=600
    )
    assert r.status_code == 200

    data = r.json()

    return data["traffic"]


def load_timers_json(zone, primary=True):
    name = zone["name"]

    # Check if the primary zone timer exists
    assert "loaded" in zone
    loaded = datetime.strptime(zone["loaded"], generic.fmt)

    if primary:
        # Check if the secondary zone timers does not exist
        assert "expires" not in zone
        assert "refresh" not in zone
        expires = None
        refresh = None
    else:
        assert "expires" in zone
        assert "refresh" in zone
        expires = datetime.strptime(zone["expires"], generic.fmt)
        refresh = datetime.strptime(zone["refresh"], generic.fmt)

    return (name, loaded, expires, refresh)


def load_zone_json(zone):
    name = zone["name"]

    return name


def test_zone_timers_primary_json(statsport):
    generic.test_zone_timers_primary(
        fetch_zones_json,
        load_timers_json,
        statsip="10.53.0.1",
        statsport=statsport,
        zonedir="ns1",
    )


def test_zone_timers_secondary_json(statsport):
    generic.test_zone_timers_secondary(
        fetch_zones_json,
        load_timers_json,
        statsip="10.53.0.3",
        statsport=statsport,
        zonedir="ns3",
    )


def test_zone_with_many_keys_json(statsport):
    generic.test_zone_with_many_keys(
        fetch_zones_json, load_zone_json, statsip="10.53.0.2", statsport=statsport
    )


def test_traffic_json(statsport):
    generic.test_traffic(fetch_traffic_json, statsip="10.53.0.2", statsport=statsport)



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import os
import select
import signal
import socket
import sys
import time

import dns.flags
import dns.message


def port():
    env_port = os.getenv("PORT")
    if env_port is None:
        env_port = 5300
    else:
        env_port = int(env_port)

    return env_port


def udp_listen(port):
    udp = socket.socket(type=socket.SOCK_DGRAM)
    udp.bind(("10.53.0.3", port))

    return udp


def tcp_listen(port):
    tcp = socket.socket(type=socket.SOCK_STREAM)
    tcp.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    tcp.bind(("10.53.0.3", port))
    tcp.listen(100)

    return tcp


def udp_tc_once(udp):
    qrybytes, clientaddr = udp.recvfrom(65535)
    qry = dns.message.from_wire(qrybytes)
    answ = dns.message.make_response(qry)
    answ.flags |= dns.flags.TC
    answbytes = answ.to_wire()
    udp.sendto(answbytes, clientaddr)


def tcp_once(tcp):
    csock, _clientaddr = tcp.accept()
    time.sleep(5)
    csock.close()


def sigterm(signum, frame):
    os.remove("ans.pid")
    sys.exit(0)


def write_pid():
    with open("ans.pid", "w") as f:
        pid = os.getpid()
        f.write("{}".format(pid))


signal.signal(signal.SIGTERM, sigterm)
write_pid()

udp = udp_listen(port())
tcp = tcp_listen(port())

input = [udp, tcp]

while True:
    try:
        inputready, outputready, exceptready = select.select(input, [], [])
    except select.error:
        break
    except socket.error:
        break
    except KeyboardInterrupt:
        break

    for s in inputready:
        if s == udp:
            udp_tc_once(udp)
        if s == tcp:
            tcp_once(tcp)

sigterm(signal.SIGTERM, 0)



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from __future__ import print_function
import os
import sys
import signal
import socket
import select
import struct
from datetime import datetime, timedelta
import time
import functools

import dns, dns.message, dns.query, dns.flags
from dns.rdatatype import *
from dns.rdataclass import *
from dns.rcode import *
from dns.name import *


# Log query to file
def logquery(type, qname):
    with open("qlog", "a") as f:
        f.write("%s %s\n", type, qname)


# Create a UDP listener
def udp_listen(ip, port, is_ipv6=False):
    try:
        udp = socket.socket(
            socket.AF_INET6 if is_ipv6 else socket.AF_INET, socket.SOCK_DGRAM
        )
        try:
            udp.bind((ip, port))
        except:
            udp.close()
            udp = None
    except:
        udp = None

    if udp is None and not is_ipv6:
        raise socket.error("Can not create an IPv4 UDP listener")

    return udp


# Create a TCP listener
def tcp_listen(ip, port, is_ipv6=False):
    try:
        tcp = socket.socket(
            socket.AF_INET6 if is_ipv6 else socket.AF_INET, socket.SOCK_STREAM
        )
        try:
            tcp.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            tcp.bind((ip, port))
            tcp.listen(100)
        except:
            tcp.close()
            tcp = None
    except:
        tcp = None

    if tcp is None and not is_ipv6:
        raise socket.error("Can not create an IPv4 TCP listener")

    return tcp


############################################################################
# Control channel - send "1" or "0" to enable or disable the "silent" mode.
############################################################################
silent = False


def ctrl_channel(msg):
    global silent

    msg = msg.splitlines().pop(0)
    print("Received control message: %s" % msg)

    if len(msg) != 1:
        return

    if silent:
        if msg == b"0":
            silent = False
            print("Silent mode was disabled")
    else:
        if msg == b"1":
            silent = True
            print("Silent mode was enabled")


############################################################################
# Respond to a DNS query.
############################################################################
def create_response(msg):
    m = dns.message.from_wire(msg)
    qname = m.question[0].name.to_text()
    rrtype = m.question[0].rdtype
    typename = dns.rdatatype.to_text(rrtype)

    with open("query.log", "a") as f:
        f.write("%s %s\n" % (typename, qname))
        print("%s %s" % (typename, qname), end=" ")

    r = dns.message.make_response(m)
    r.set_rcode(NOERROR)
    if rrtype == A:
        tld = qname.split(".")[-2] + "."
        ns = "local." + tld
        r.answer.append(dns.rrset.from_text(qname, 300, IN, A, "10.53.0.11"))
        r.answer.append(dns.rrset.from_text(tld, 300, IN, NS, "local." + tld))
        r.additional.append(dns.rrset.from_text(ns, 300, IN, A, "10.53.0.11"))
    elif rrtype == NS:
        r.answer.append(dns.rrset.from_text(qname, 300, IN, NS, "."))
    elif rrtype == SOA:
        r.answer.append(dns.rrset.from_text(qname, 300, IN, SOA, ". . 0 0 0 0 0"))
    else:
        r.authority.append(dns.rrset.from_text(qname, 300, IN, SOA, ". . 0 0 0 0 0"))
    r.flags |= dns.flags.AA
    return r


def sigterm(signum, frame):
    print("Shutting down now...")
    os.remove("ans.pid")
    running = False
    sys.exit(0)


############################################################################
# Main
#
# Set up responder and control channel, open the pid file, and start
# the main loop, listening for queries on the query channel or commands
# on the control channel and acting on them.
############################################################################
ip4 = "10.53.0.11"
ip6 = "fd92:7065:b8e:ffff::11"

try:
    port = int(os.environ["PORT"])
except:
    port = 5300

try:
    ctrlport = int(os.environ["EXTRAPORT1"])
except:
    ctrlport = 5300

ctrl4_tcp = tcp_listen(ip4, ctrlport)
query4_udp = udp_listen(ip4, port)
query6_udp = udp_listen(ip6, port, is_ipv6=True)
query4_tcp = tcp_listen(ip4, port)
query6_tcp = tcp_listen(ip6, port, is_ipv6=True)

havev6 = query6_udp is not None and query6_tcp is not None

signal.signal(signal.SIGTERM, sigterm)

f = open("ans.pid", "w")
pid = os.getpid()
print(pid, file=f)
f.close()

running = True

print("Listening on %s port %d" % (ip4, ctrlport))
print("Listening on %s port %d" % (ip4, port))
if havev6:
    print("Listening on %s port %d" % (ip6, port))

print("Ctrl-c to quit")

if havev6:
    input = [ctrl4_tcp, query4_udp, query6_udp, query4_tcp, query6_tcp]
else:
    input = [ctrl4_tcp, query4_udp, query4_tcp]

hung_conns = []

while running:
    try:
        inputready, outputready, exceptready = select.select(input, [], [])
    except select.error as e:
        break
    except socket.error as e:
        break
    except KeyboardInterrupt:
        break

    for s in inputready:
        if s == ctrl4_tcp:
            print("Control channel connected")
            conn = None
            try:
                # Handle control channel input
                conn, addr = s.accept()
                msg = conn.recv(1)
                if msg:
                    ctrl_channel(msg)
                conn.close()
            except s.timeout:
                pass
            if conn:
                conn.close()
        elif s == query4_tcp or s == query6_tcp:
            print(
                "TCP query received on %s" % (ip4 if s == query4_tcp else ip6), end=" "
            )
            conn = None
            try:
                # Handle incoming queries
                conn, addr = s.accept()
                if not silent:
                    # get TCP message length
                    msg = conn.recv(2)
                    if len(msg) != 2:
                        print("NO RESPONSE (can not read the message length)")
                        conn.close()
                        continue
                    length = struct.unpack(">H", msg[:2])[0]
                    msg = conn.recv(length)
                    if len(msg) != length:
                        print("NO RESPONSE (can not read the message)")
                        conn.close()
                        continue
                    rsp = create_response(msg)
                    if rsp:
                        print(dns.rcode.to_text(rsp.rcode()))
                        wire = rsp.to_wire()
                        conn.send(struct.pack(">H", len(wire)))
                        conn.send(wire)
                    else:
                        print("NO RESPONSE (can not create a response)")
                else:
                    # Do not respond and hang the connection.
                    print("NO RESPONSE (silent mode)")
                    hung_conns.append(conn)
                    continue
            except socket.error as e:
                print("NO RESPONSE (error: %s)" % str(e))
            if conn:
                conn.close()
        elif s == query4_udp or s == query6_udp:
            print(
                "UDP query received on %s" % (ip4 if s == query4_udp else ip6), end=" "
            )
            # Handle incoming queries
            msg = s.recvfrom(65535)
            if not silent:
                rsp = create_response(msg[0])
                if rsp:
                    print(dns.rcode.to_text(rsp.rcode()))
                    s.sendto(rsp.to_wire(), msg[1])
                else:
                    print("NO RESPONSE (can not create a response)")
            else:
                # Do not respond.
                print("NO RESPONSE (silent mode)")
    if not running:
        break



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import os
from typing import Optional

import dns.query
import dns.message


QUERY_TIMEOUT = 10


def udp(
    message: dns.message.Message, ip: str, port: Optional[int] = None
) -> dns.message.Message:
    if port is None:
        port = int(os.environ["PORT"])
    return dns.query.udp(message, ip, QUERY_TIMEOUT, port=port)


def tcp(
    message: dns.message.Message, ip: str, port: Optional[int] = None
) -> dns.message.Message:
    if port is None:
        port = int(os.environ["PORT"])
    return dns.query.tcp(message, ip, QUERY_TIMEOUT, port=port)



#!/usr/bin/env python

# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import os
import sys
import socket
import subprocess
import random
import time

from functools import reduce
from resource import getrlimit
from resource import setrlimit
from resource import RLIMIT_NOFILE

MULTIDIG_INSTANCES = 10
CONNECT_TRIES = 5

random.seed()

# Ensure we have enough file desriptors to work
rlimit_nofile = getrlimit(RLIMIT_NOFILE)
if rlimit_nofile[0] < 1024:
    setrlimit(RLIMIT_NOFILE, (1024, rlimit_nofile[1]))


# Introduce some random delay
def jitter():
    time.sleep((500 + random.randint(0, 250)) / 1000000.0)


# A set of simple procedures to get the test's configuration options
def get_http_port(http_secure=False):
    http_port_env = None
    if http_secure:
        http_port_env = os.getenv("HTTPSPORT")
    else:
        http_port_env = os.getenv("HTTPPORT")
    if http_port_env:
        return int(http_port_env)
    return 443


def get_http_host():
    bind_host = os.getenv("BINDHOST")
    if bind_host:
        return bind_host
    return "localhost"


def get_dig_path():
    dig_path = os.getenv("DIG")
    if dig_path:
        return dig_path
    return "dig"


# A simple class which creates the given number of TCP connections to
# the given host in order to stress the BIND's quota facility
class TCPConnector:
    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.connections = []

    def connect_one(self):
        tries = CONNECT_TRIES
        while tries > 0:
            try:
                sock = socket.create_connection(
                    address=(self.host, self.port), timeout=None
                )
                self.connections.append(sock)
                break
            except ConnectionResetError:
                # some jitter for BSDs
                jitter()
                continue
            except TimeoutError:
                jitter()
                continue
            finally:
                tries -= 1

    # Close an established connection (randomly)
    def disconnect_random(self):
        pos = random.randint(0, len(self.connections) - 1)
        conn = self.connections[pos]
        try:
            conn.shutdown(socket.SHUT_RDWR)
            conn.close()
        except OSError:
            conn.close()
        finally:
            self.connections.remove(conn)

    def disconnect_all(self):
        while len(self.connections) != 0:
            self.disconnect_random()


# A simple class which allows running a dig instance under control of
# the process
class SubDIG:
    def __init__(self, http_secure=None, extra_args=None):
        self.sub_process = None
        self.dig_path = get_dig_path()
        self.host = get_http_host()
        self.port = get_http_port(http_secure=http_secure)
        if http_secure:
            self.http_secure = True
        else:
            self.http_secure = False
        self.extra_args = extra_args

    # This method constructs a command string
    def get_command(self):
        command = self.dig_path + " -p " + str(self.port) + " "
        command = command + "+noadd +nosea +nostat +noquest +nocmd +time=30 "
        if self.http_secure:
            command = command + "+https "
        else:
            command = command + "+http-plain "
        command = command + "@" + self.host + " "
        if self.extra_args:
            command = command + self.extra_args
        return command

    def run(self):
        # pylint: disable=consider-using-with
        with open(os.devnull, "w", encoding="utf-8") as devnull:
            self.sub_process = subprocess.Popen(
                self.get_command(), shell=True, stdout=devnull
            )

    def wait(self, timeout=None):
        res = None
        if timeout is None:
            return self.sub_process.wait()
        try:
            res = self.sub_process.wait(timeout=timeout)
        except subprocess.TimeoutExpired:
            return None
        return res

    def alive(self):
        return self.sub_process.poll() is None


# A simple wrapper class which allows running multiple dig instances
# and examining their statuses in one logical operation.
class MultiDIG:
    def __init__(self, numdigs, http_secure=None, extra_args=None):
        assert int(numdigs) > 0, f"numdigs={numdigs}"
        digs = []
        for _ in range(1, int(numdigs) + 1):
            digs.append(SubDIG(http_secure=http_secure, extra_args=extra_args))
        self.digs = digs
        assert len(self.digs) == int(numdigs), f"len={len(self.digs)} numdigs={numdigs}"

    def run(self):
        for p in self.digs:
            p.run()

    def wait(self):
        return map(lambda p: (p.wait()), self.digs)

    # Wait for the all instances to terminate with expected given
    # status. Returns true or false.
    def wait_for_result(self, result):
        return reduce(
            lambda a, b: ((a == result or a is True) and b == result), self.wait()
        )

    def alive(self):
        return reduce(lambda a, b: (a and b), map(lambda p: (p.alive()), self.digs))

    def completed(self):
        total = 0
        for p in self.digs:
            if not p.alive():
                total += 1
        return total


# The test's main logic
def run_test(http_secure=True):
    query_args = "SOA ."
    # Let's try to make a successful query
    subdig = SubDIG(http_secure=http_secure, extra_args=query_args)
    subdig.run()
    assert subdig.wait() == 0, "DIG was expected to succeed"
    # Let's create a lot of TCP connections to the server stress the
    # HTTP quota
    connector = TCPConnector(get_http_host(), get_http_port(http_secure=http_secure))
    # Let's make queries until the quota kicks in
    subdig = SubDIG(http_secure=http_secure, extra_args=query_args)
    subdig.run()
    while True:
        connector.connect_one()
        subdig = SubDIG(http_secure=http_secure, extra_args=query_args)
        subdig.run()
        if subdig.wait(timeout=5) is None:
            break

    # At this point quota has kicked in.  Additionally, let's create a
    # bunch of dig processes all trying to make a query against the
    # server with exceeded quota
    multidig = MultiDIG(
        MULTIDIG_INSTANCES, http_secure=http_secure, extra_args=query_args
    )
    multidig.run()
    # Wait for the dig instance to complete. Not a single instance has
    # a chance to complete successfully because of the exceeded quota
    assert (
        subdig.wait(timeout=5) is None
    ), "The single DIG instance has stopped prematurely"
    assert subdig.alive(), "The single DIG instance is expected to be alive"
    assert multidig.alive(), (
        "The DIG instances from the set are all expected to "
        "be alive, but {} of them have completed"
    ).format(multidig.completed())
    # Let's close opened connections (in random order) to let all dig
    # processes to complete
    connector.disconnect_all()
    # Wait for all processes to complete successfully
    assert subdig.wait() == 0, "Single DIG instance failed"
    assert (
        multidig.wait_for_result(0) is True
    ), "One or more of DIG instances returned unexpected results"


def main():
    run_test(http_secure=True)
    run_test(http_secure=False)
    # If we have reached this point we could safely return 0
    # (success). If the test fails because of an assert, the whole
    # program will return non-zero exit code and produce the backtrace
    return 0


sys.exit(main())



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import concurrent.futures
import os
import subprocess
import time
import dns.query
import dns.update


def rndc_loop(test_state, server):
    rndc = os.getenv("RNDC")
    port = os.getenv("CONTROLPORT")

    cmdline = [
        rndc,
        "-c",
        "../_common/rndc.conf",
        "-p",
        port,
        "-s",
        server,
        "reload",
    ]

    while not test_state["finished"]:
        subprocess.run(cmdline, check=False)
        time.sleep(1)


def update_zone(test_state, zone, named_port, logger):
    server = "10.53.0.2"
    for i in range(1000):
        if test_state["finished"]:
            return
        update = dns.update.UpdateMessage(zone)
        update.add(f"dynamic-{i}.{zone}", 300, "TXT", f"txt-{i}")
        try:
            response = dns.query.udp(update, server, 10, named_port)
            assert response.rcode() == dns.rcode.NOERROR
        except dns.exception.Timeout:
            logger.info(f"error: query timeout for {zone}")

    logger.info(f"Update of {server} zone {zone} successful")


# If the test has run to completion without named crashing, it has succeeded.
def test_update_stress(named_port, logger):
    test_state = {"finished": False}

    with concurrent.futures.ThreadPoolExecutor() as executor:
        executor.submit(rndc_loop, test_state, "10.53.0.3")

        updaters = []
        for i in range(5):
            zone = f"zone00000{i}.example."
            updaters.append(
                executor.submit(update_zone, test_state, zone, named_port, logger)
            )

        # All the update_zone() tasks are expected to complete within 5
        # minutes.  If they do not, we cannot assert immediately as that will
        # cause the ThreadPoolExecutor context manager to wait indefinitely;
        # instead, we first signal all tasks that it is time to exit and only
        # check whether any task failed to finish within 5 minutes outside of
        # the ThreadPoolExecutor context manager.
        unfinished_tasks = concurrent.futures.wait(updaters, timeout=300).not_done
        test_state["finished"] = True

    assert not unfinished_tasks



#!/usr/bin/env python
#
# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# Convert automake .trs files into JUnit format suitable for Gitlab

import argparse
import os
import sys
from xml.etree import ElementTree
from xml.etree.ElementTree import Element
from xml.etree.ElementTree import SubElement


# getting explicit encoding specification right for Python 2/3 would be messy,
# so let's hope for the best
def read_whole_text(filename):
    with open(filename) as inf:  # pylint: disable-msg=unspecified-encoding
        return inf.read().strip()


def read_trs_result(filename):
    result = None
    with open(filename, "r") as trs:  # pylint: disable-msg=unspecified-encoding
        for line in trs:
            items = line.split()
            if len(items) < 2:
                raise ValueError("unsupported line in trs file", filename, line)
            if items[0] != (":test-result:"):
                continue
            if result is not None:
                raise NotImplementedError("double :test-result:", filename)
            result = items[1].upper()

    if result is None:
        raise ValueError(":test-result: not found", filename)

    return result


def find_test_relative_path(source_dir, in_path):
    """Return {in_path}.c if it exists, with fallback to {in_path}"""
    candidates_relative = [in_path + ".c", in_path]
    for relative in candidates_relative:
        absolute = os.path.join(source_dir, relative)
        if os.path.exists(absolute):
            return relative
    raise KeyError


def err_out(exception):
    raise exception


def walk_trss(source_dir):
    for cur_dir, _dirs, files in os.walk(source_dir, onerror=err_out):
        for filename in files:
            if not filename.endswith(".trs"):
                continue

            filename_prefix = filename[: -len(".trs")]
            log_name = filename_prefix + ".log"
            full_trs_path = os.path.join(cur_dir, filename)
            full_log_path = os.path.join(cur_dir, log_name)
            sub_dir = os.path.relpath(cur_dir, source_dir)
            test_name = os.path.join(sub_dir, filename_prefix)

            t = {
                "name": test_name,
                "full_log_path": full_log_path,
                "rel_log_path": os.path.relpath(full_log_path, source_dir),
            }
            t["result"] = read_trs_result(full_trs_path)

            # try to find dir/file path for a clickable link
            try:
                t["rel_file_path"] = find_test_relative_path(source_dir, test_name)
            except KeyError:
                pass  # no existing path found

            yield t


def append_testcase(testsuite, t):
    # attributes taken from
    # https://gitlab.com/gitlab-org/gitlab-foss/-/blob/master/lib/gitlab/ci/parsers/test/junit.rb
    attrs = {"name": t["name"]}
    if "rel_file_path" in t:
        attrs["file"] = t["rel_file_path"]

    testcase = SubElement(testsuite, "testcase", attrs)

    # Gitlab accepts only [[ATTACHMENT| links for system-out, not raw text
    s = SubElement(testcase, "system-out")
    s.text = "[[ATTACHMENT|" + t["rel_log_path"] + "]]"
    if t["result"].lower() == "pass":
        return

    # Gitlab shows output only for failed or skipped tests
    if t["result"].lower() == "skip":
        err = SubElement(testcase, "skipped")
    else:
        err = SubElement(testcase, "failure")
    err.text = read_whole_text(t["full_log_path"])


def gen_junit(results):
    testsuites = Element("testsuites")
    testsuite = SubElement(testsuites, "testsuite")
    for test in results:
        append_testcase(testsuite, test)
    return testsuites


def check_directory(path):
    try:
        os.listdir(path)
        return path
    except OSError as ex:
        msg = "Path {} cannot be listed as a directory: {}".format(path, ex)
        raise argparse.ArgumentTypeError(msg)


def main():
    parser = argparse.ArgumentParser(
        description="Recursively search for .trs + .log files and compile "
        "them into JUnit XML suitable for Gitlab. Paths in the "
        "XML are relative to the specified top directory."
    )
    parser.add_argument(
        "top_directory",
        type=check_directory,
        help="root directory where to start scanning for .trs files",
    )
    args = parser.parse_args()
    junit = gen_junit(walk_trss(args.top_directory))

    # encode results into file format, on Python 3 it produces bytes
    xml = ElementTree.tostring(junit, "utf-8")
    # use stdout as a binary file object, Python2/3 compatibility
    output = getattr(sys.stdout, "buffer", sys.stdout)
    output.write(xml)


if __name__ == "__main__":
    main()



#!/usr/bin/python3

# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import os
import pathlib
import subprocess

import pytest


def is_pid_alive(pid):
    try:
        os.kill(pid, 0)
        return True
    except OSError:
        return False


def run_sslyze_in_a_loop(executable, port, log_file_prefix):
    # Determine the PID of ns1.
    with open(pathlib.Path("ns1", "named.pid"), encoding="utf-8") as pidfile:
        pid = int(pidfile.read())

    # Ensure ns1 is alive before starting the loop below to avoid reporting
    # false positives.
    if not is_pid_alive(pid):
        pytest.skip(f"ns1 (PID: {pid}) is not running")

    # Run sslyze on ns1 in a loop with a limit of 30 iterations.  Interrupt the
    # test as soon as ns1 is determined to not be running any more.  Log sslyze
    # output.
    sslyze_args = [executable, f"10.53.0.1:{port}"]
    for i in range(0, 30):
        log_file = f"{log_file_prefix}.ns1.{port}.{i + 1}"
        with open(log_file, "wb") as sslyze_log:
            # Run sslyze, logging stdout+stderr.  Ignore the exit code since
            # sslyze is only used for triggering crashes here rather than
            # actual TLS analysis.
            subprocess.run(
                sslyze_args,
                stdout=sslyze_log,
                stderr=subprocess.STDOUT,
                timeout=30,
                check=False,
            )
            # Ensure ns1 is still alive after each sslyze run.
            assert is_pid_alive(pid), f"ns1 (PID: {pid}) exited prematurely"


def test_sslyze_doh(sslyze_executable, named_httpsport):
    run_sslyze_in_a_loop(sslyze_executable, named_httpsport, "sslyze.log.doh")


def test_sslyze_dot(sslyze_executable, named_tlsport):
    run_sslyze_in_a_loop(sslyze_executable, named_tlsport, "sslyze.log.dot")



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from functools import partial
import logging
import os
from pathlib import Path
import re
import shutil
import subprocess
import tempfile
import time
from typing import Any, Dict, List, Optional

import pytest


pytest.register_assert_rewrite("isctest")


# Silence warnings caused by passing a pytest fixture to another fixture.
# pylint: disable=redefined-outer-name


# ----------------- Older pytest / xdist compatibility -------------------
# As of 2023-01-11, the minimal supported pytest / xdist versions are
# determined by what is available in EL8/EPEL8:
# - pytest 3.4.2
# - pytest-xdist 1.24.1
_pytest_ver = pytest.__version__.split(".")
_pytest_major_ver = int(_pytest_ver[0])
if _pytest_major_ver < 7:
    # pytest.Stash/pytest.StashKey mechanism has been added in 7.0.0
    # for older versions, use regular dictionary with string keys instead
    FIXTURE_OK = "fixture_ok"  # type: Any
else:
    FIXTURE_OK = pytest.StashKey[bool]()  # pylint: disable=no-member

# ----------------------- Globals definition -----------------------------

LOG_FORMAT = "%(asctime)s %(levelname)7s:%(name)s  %(message)s"
XDIST_WORKER = os.environ.get("PYTEST_XDIST_WORKER", "")
FILE_DIR = os.path.abspath(Path(__file__).parent)
ENV_RE = re.compile(b"([^=]+)=(.*)")
PORT_MIN = 5001
PORT_MAX = 32767
PORTS_PER_TEST = 20
PRIORITY_TESTS = [
    # Tests that are scheduled first. Speeds up parallel execution.
    "dupsigs/",
    "rpz/",
    "rpzrecurse/",
    "serve-stale/",
    "timeouts/",
    "upforwd/",
]
PRIORITY_TESTS_RE = re.compile("|".join(PRIORITY_TESTS))
CONFTEST_LOGGER = logging.getLogger("conftest")
SYSTEM_TEST_DIR_GIT_PATH = "bin/tests/system"
SYSTEM_TEST_NAME_RE = re.compile(f"{SYSTEM_TEST_DIR_GIT_PATH}" + r"/([^/]+)")
SYMLINK_REPLACEMENT_RE = re.compile(r"/tests(_.*)\.py")

# ---------------------- Module initialization ---------------------------


def init_pytest_conftest_logger(conftest_logger):
    """
    This initializes the conftest logger which is used for pytest setup
    and configuration before tests are executed -- aka any logging in this
    file that is _not_ module-specific.
    """
    conftest_logger.setLevel(logging.DEBUG)
    file_handler = logging.FileHandler("pytest.conftest.log.txt")
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(logging.Formatter(LOG_FORMAT))
    conftest_logger.addHandler(file_handler)


init_pytest_conftest_logger(CONFTEST_LOGGER)


def avoid_duplicated_logs():
    """
    Remove direct root logger output to file descriptors.
    This default is causing duplicates because all our messages go through
    regular logging as well and are thus displayed twice.
    """
    todel = []
    for handler in logging.root.handlers:
        if handler.__class__ == logging.StreamHandler:
            # Beware: As for pytest 7.2.2, LiveLogging and LogCapture
            # handlers inherit from logging.StreamHandler
            todel.append(handler)
    for handler in todel:
        logging.root.handlers.remove(handler)


def parse_env(env_bytes):
    """Parse the POSIX env format into Python dictionary."""
    out = {}
    for line in env_bytes.splitlines():
        match = ENV_RE.match(line)
        if match:
            # EL8+ workaround for https://access.redhat.com/solutions/6994985
            # FUTURE: can be removed when we no longer need to parse env vars
            if match.groups()[0] in [b"which_declare", b"BASH_FUNC_which%%"]:
                continue
            out[match.groups()[0]] = match.groups()[1]
    return out


def get_env_bytes(cmd):
    try:
        proc = subprocess.run(
            [cmd],
            shell=True,
            check=True,
            cwd=FILE_DIR,
            stdout=subprocess.PIPE,
        )
    except subprocess.CalledProcessError as exc:
        CONFTEST_LOGGER.error("failed to get shell env: %s", exc)
        raise exc
    env_bytes = proc.stdout
    return parse_env(env_bytes)


# Read common environment variables for running tests from conf.sh.
# FUTURE: Remove conf.sh entirely and define all variables in pytest only.
CONF_ENV = get_env_bytes(". ./conf.sh && env")
os.environb.update(CONF_ENV)
CONFTEST_LOGGER.debug("variables in env: %s", ", ".join([str(key) for key in CONF_ENV]))

# --------------------------- pytest hooks -------------------------------


def pytest_addoption(parser):
    parser.addoption(
        "--noclean",
        action="store_true",
        default=False,
        help="don't remove the temporary test directories with artifacts",
    )


def pytest_configure(config):
    # Ensure this hook only runs on the main pytest instance if xdist is
    # used to spawn other workers.
    if not XDIST_WORKER:
        if config.pluginmanager.has_plugin("xdist") and config.option.numprocesses:
            # system tests depend on module scope for setup & teardown
            # enforce use "loadscope" scheduler or disable paralelism
            try:
                import xdist.scheduler.loadscope  # pylint: disable=unused-import
            except ImportError:
                CONFTEST_LOGGER.debug(
                    "xdist is too old and does not have "
                    "scheduler.loadscope, disabling parallelism"
                )
                config.option.dist = "no"
            else:
                config.option.dist = "loadscope"


def pytest_ignore_collect(path):
    # System tests are executed in temporary directories inside
    # bin/tests/system. These temporary directories contain all files
    # needed for the system tests - including tests_*.py files. Make sure to
    # ignore these during test collection phase. Otherwise, test artifacts
    # from previous runs could mess with the runner. Also ignore the
    # convenience symlinks to those test directories. In both of those
    # cases, the system test name (directory) contains an underscore, which
    # is otherwise and invalid character for a system test name.
    match = SYSTEM_TEST_NAME_RE.search(str(path))
    if match is None:
        CONFTEST_LOGGER.warning("unexpected test path: %s (ignored)", path)
        return True
    system_test_name = match.groups()[0]
    return "_" in system_test_name


def pytest_collection_modifyitems(items):
    """Schedule long-running tests first to get more benefit from parallelism."""
    priority = []
    other = []
    for item in items:
        if PRIORITY_TESTS_RE.search(item.nodeid):
            priority.append(item)
        else:
            other.append(item)
    items[:] = priority + other


class NodeResult:
    def __init__(self, report=None):
        self.outcome = None
        self.messages = []
        if report is not None:
            self.update(report)

    def update(self, report):
        if self.outcome is None or report.outcome != "passed":
            self.outcome = report.outcome
        if report.longreprtext:
            self.messages.append(report.longreprtext)


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item):
    """Hook that is used to expose test results to session (for use in fixtures)."""
    # execute all other hooks to obtain the report object
    outcome = yield
    report = outcome.get_result()

    # Set the test outcome in session, so we can access it from module-level
    # fixture using nodeid. Note that this hook is called three times: for
    # setup, call and teardown. We only care about the overall result so we
    # merge the results together and preserve the information whether a test
    # passed.
    test_results = {}
    try:
        test_results = getattr(item.session, "test_results")
    except AttributeError:
        setattr(item.session, "test_results", test_results)
    node_result = test_results.setdefault(item.nodeid, NodeResult())
    node_result.update(report)


# --------------------------- Fixtures -----------------------------------


@pytest.fixture(scope="session")
def modules():
    """
    Sorted list of ALL modules.

    The list includes even test modules that are not tested in the current
    session. It is used to determine port distribution. Using a complete
    list of all possible test modules allows independent concurrent pytest
    invocations.
    """
    mods = []
    for dirpath, _dirs, files in os.walk(FILE_DIR):
        for file in files:
            if file.startswith("tests_") and file.endswith(".py"):
                mod = f"{dirpath}/{file}"
                if not pytest_ignore_collect(mod):
                    mods.append(mod)
    return sorted(mods)


@pytest.fixture(scope="session")
def module_base_ports(modules):
    """
    Dictionary containing assigned base port for every module.

    The port numbers are deterministically assigned before any testing
    starts. This fixture MUST return the same value when called again
    during the same test session. When running tests in parallel, this is
    exactly what happens - every worker thread will call this fixture to
    determine test ports.
    """
    port_min = PORT_MIN
    port_max = PORT_MAX - len(modules) * PORTS_PER_TEST
    if port_max < port_min:
        raise RuntimeError("not enough ports to assign unique port set to each module")

    # Rotate the base port value over time to detect possible test issues
    # with using random ports. This introduces a very slight race condition
    # risk. If this value changes between pytest invocation and spawning
    # worker threads, multiple tests may have same port values assigned. If
    # these tests are then executed simultaneously, the test results will
    # be misleading.
    base_port = int(time.time() // 3600) % (port_max - port_min) + port_min

    return {mod: base_port + i * PORTS_PER_TEST for i, mod in enumerate(modules)}


@pytest.fixture(scope="module")
def base_port(request, module_base_ports):
    """Start of the port range assigned to a particular test module."""
    port = module_base_ports[request.fspath]
    return port


@pytest.fixture(scope="module")
def ports(base_port):
    """Dictionary containing port names and their assigned values."""
    return {
        "PORT": base_port,
        "TLSPORT": base_port + 1,
        "HTTPPORT": base_port + 2,
        "HTTPSPORT": base_port + 3,
        "EXTRAPORT1": base_port + 4,
        "EXTRAPORT2": base_port + 5,
        "EXTRAPORT3": base_port + 6,
        "EXTRAPORT4": base_port + 7,
        "EXTRAPORT5": base_port + 8,
        "EXTRAPORT6": base_port + 9,
        "EXTRAPORT7": base_port + 10,
        "EXTRAPORT8": base_port + 11,
        "CONTROLPORT": base_port + 12,
    }


@pytest.fixture(scope="module")
def named_port(ports):
    return ports["PORT"]


@pytest.fixture(scope="module")
def named_tlsport(ports):
    return ports["TLSPORT"]


@pytest.fixture(scope="module")
def named_httpsport(ports):
    return ports["HTTPSPORT"]


@pytest.fixture(scope="module")
def control_port(ports):
    return ports["CONTROLPORT"]


@pytest.fixture(scope="module")
def env(ports):
    """Dictionary containing environment variables for the test."""
    env = os.environ.copy()
    for portname, portnum in ports.items():
        env[portname] = str(portnum)
    env["builddir"] = f"{env['TOP_BUILDDIR']}/{SYSTEM_TEST_DIR_GIT_PATH}"
    env["srcdir"] = f"{env['TOP_SRCDIR']}/{SYSTEM_TEST_DIR_GIT_PATH}"
    return env


@pytest.fixture(scope="module")
def system_test_name(request):
    """Name of the system test directory."""
    path = Path(request.fspath)
    return path.parent.name


@pytest.fixture(scope="module")
def mlogger(system_test_name):
    """Logging facility specific to this test module."""
    avoid_duplicated_logs()
    return logging.getLogger(system_test_name)


@pytest.fixture
def logger(request, system_test_name):
    """Logging facility specific to a particular test."""
    return logging.getLogger(f"{system_test_name}.{request.node.name}")


@pytest.fixture(scope="module")
def system_test_dir(
    request, env, system_test_name, mlogger
):  # pylint: disable=too-many-statements,too-many-locals
    """
    Temporary directory for executing the test.

    This fixture is responsible for creating (and potentially removing) a
    copy of the system test directory which is used as a temporary
    directory for the test execution.

    FUTURE: This removes the need to have clean.sh scripts.
    """

    def get_test_result():
        """Aggregate test results from all individual tests from this module
        into a single result: failed > skipped > passed."""
        try:
            all_test_results = request.session.test_results
        except AttributeError:
            # This may happen if pytest execution is interrupted and
            # pytest_runtest_makereport() is never called.
            mlogger.debug("can't obtain test results, test run was interrupted")
            return "error"
        test_results = {
            node.nodeid: all_test_results[node.nodeid]
            for node in request.node.collect()
            if node.nodeid in all_test_results
        }
        assert len(test_results)
        messages = []
        for node, result in test_results.items():
            mlogger.debug("%s %s", result.outcome.upper(), node)
            messages.extend(result.messages)
        for message in messages:
            mlogger.debug("\n" + message)
        failed = any(res.outcome == "failed" for res in test_results.values())
        skipped = any(res.outcome == "skipped" for res in test_results.values())
        if failed:
            return "failed"
        if skipped:
            return "skipped"
        assert all(res.outcome == "passed" for res in test_results.values())
        return "passed"

    def unlink(path):
        try:
            path.unlink()  # missing_ok=True isn't available on Python 3.6
        except FileNotFoundError:
            pass

    # Create a temporary directory with a copy of the original system test dir contents
    system_test_root = Path(f"{env['TOP_BUILDDIR']}/{SYSTEM_TEST_DIR_GIT_PATH}")
    testdir = Path(
        tempfile.mkdtemp(prefix=f"{system_test_name}_tmp_", dir=system_test_root)
    )
    shutil.rmtree(testdir)
    shutil.copytree(system_test_root / system_test_name, testdir)

    # Create a convenience symlink with a stable and predictable name
    module_name = SYMLINK_REPLACEMENT_RE.sub(r"\1", request.node.name)
    symlink_dst = system_test_root / module_name
    unlink(symlink_dst)
    symlink_dst.symlink_to(os.path.relpath(testdir, start=system_test_root))

    # Configure logger to write to a file inside the temporary test directory
    mlogger.handlers.clear()
    mlogger.setLevel(logging.DEBUG)
    handler = logging.FileHandler(testdir / "pytest.log.txt", mode="w")
    formatter = logging.Formatter(LOG_FORMAT)
    handler.setFormatter(formatter)
    mlogger.addHandler(handler)

    # System tests are meant to be executed from their directory - switch to it.
    old_cwd = os.getcwd()
    os.chdir(testdir)
    mlogger.debug("switching to tmpdir: %s", testdir)
    try:
        yield testdir  # other fixtures / tests will execute here
    finally:
        os.chdir(old_cwd)
        mlogger.debug("changed workdir to: %s", old_cwd)

        result = get_test_result()

        # Clean temporary dir unless it should be kept
        keep = False
        if request.config.getoption("--noclean"):
            mlogger.debug(
                "--noclean requested, keeping temporary directory %s", testdir
            )
            keep = True
        elif result == "failed":
            mlogger.debug(
                "test failure detected, keeping temporary directory %s", testdir
            )
            keep = True
        elif not request.node.stash[FIXTURE_OK]:
            mlogger.debug(
                "test setup/teardown issue detected, keeping temporary directory %s",
                testdir,
            )
            keep = True

        if keep:
            mlogger.info(
                "test artifacts in: %s", symlink_dst.relative_to(system_test_root)
            )
        else:
            mlogger.debug("deleting temporary directory")
            handler.flush()
            handler.close()
            shutil.rmtree(testdir)
            unlink(symlink_dst)


def _run_script(  # pylint: disable=too-many-arguments
    env,
    mlogger,
    system_test_dir: Path,
    interpreter: str,
    script: str,
    args: Optional[List[str]] = None,
):
    """Helper function for the shell / perl script invocations (through fixtures below)."""
    if args is None:
        args = []
    path = Path(script)
    if not path.is_absolute():
        # make sure relative paths are always relative to system_dir
        path = system_test_dir.parent / path
    script = str(path)
    cwd = os.getcwd()
    if not path.exists():
        raise FileNotFoundError(f"script {script} not found in {cwd}")
    mlogger.debug("running script: %s %s %s", interpreter, script, " ".join(args))
    mlogger.debug("  workdir: %s", cwd)
    returncode = 1

    cmd = [interpreter, script] + args
    with subprocess.Popen(
        cmd,
        env=env,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        universal_newlines=True,
        errors="backslashreplace",
    ) as proc:
        if proc.stdout:
            for line in proc.stdout:
                mlogger.info("    %s", line.rstrip("\n"))
        proc.communicate()
        returncode = proc.returncode
        if returncode:
            raise subprocess.CalledProcessError(returncode, cmd)
        mlogger.debug("  exited with %d", returncode)


@pytest.fixture(scope="module")
def shell(env, system_test_dir, mlogger):
    """Function to call a shell script with arguments."""
    return partial(_run_script, env, mlogger, system_test_dir, env["SHELL"])


@pytest.fixture(scope="module")
def perl(env, system_test_dir, mlogger):
    """Function to call a perl script with arguments."""
    return partial(_run_script, env, mlogger, system_test_dir, env["PERL"])


@pytest.fixture(scope="module")
def run_tests_sh(system_test_dir, shell):
    """Utility function to execute tests.sh as a python test."""

    def run_tests():
        shell(f"{system_test_dir}/tests.sh")

    return run_tests


@pytest.fixture(scope="module", autouse=True)
def system_test(  # pylint: disable=too-many-arguments,too-many-statements
    request,
    env: Dict[str, str],
    mlogger,
    system_test_dir,
    shell,
    perl,
):
    """
    Driver of the test setup/teardown process. Used automatically for every test module.

    This is the most important one-fixture-to-rule-them-all. Note the
    autouse=True which causes this fixture to be loaded by every test
    module without the need to explicitly specify it.

    When this fixture is used, it utilizes other fixtures, such as
    system_test_dir, which handles the creation of the temporary test
    directory.

    Afterwards, it checks the test environment and takes care of starting
    the servers. When everything is ready, that's when the actual tests are
    executed. Once that is done, this fixture stops the servers and checks
    for any artifacts indicating an issue (e.g. coredumps).

    Finally, when this fixture reaches an end (or encounters an exception,
    which may be caused by fail/skip invocations), any fixtures which is
    used by this one are finalized - e.g. system_test_dir performs final
    checks and cleans up the temporary test directory.
    """

    def check_net_interfaces():
        try:
            perl("testsock.pl", ["-p", env["PORT"]])
        except subprocess.CalledProcessError as exc:
            mlogger.error("testsock.pl: exited with code %d", exc.returncode)
            pytest.skip("Network interface aliases not set up.")

    def check_prerequisites():
        try:
            shell(f"{system_test_dir}/prereq.sh")
        except FileNotFoundError:
            pass  # prereq.sh is optional
        except subprocess.CalledProcessError:
            pytest.skip("Prerequisites missing.")

    def setup_test():
        try:
            shell(f"{system_test_dir}/setup.sh")
        except FileNotFoundError:
            pass  # setup.sh is optional
        except subprocess.CalledProcessError as exc:
            mlogger.error("Failed to run test setup")
            pytest.fail(f"setup.sh exited with {exc.returncode}")

    def start_servers():
        try:
            perl("start.pl", ["--port", env["PORT"], system_test_dir.name])
        except subprocess.CalledProcessError as exc:
            mlogger.error("Failed to start servers")
            pytest.fail(f"start.pl exited with {exc.returncode}")

    def stop_servers():
        try:
            perl("stop.pl", [system_test_dir.name])
        except subprocess.CalledProcessError as exc:
            mlogger.error("Failed to stop servers")
            get_core_dumps()
            pytest.fail(f"stop.pl exited with {exc.returncode}")

    def get_core_dumps():
        try:
            shell("get_core_dumps.sh", [system_test_dir.name])
        except subprocess.CalledProcessError as exc:
            mlogger.error("Found core dumps or sanitizer reports")
            pytest.fail(f"get_core_dumps.sh exited with {exc.returncode}")

    os.environ.update(env)  # Ensure pytests have the same env vars as shell tests.
    mlogger.info(f"test started: {request.node.name}")
    port = int(env["PORT"])
    mlogger.info("using port range: <%d, %d>", port, port + PORTS_PER_TEST - 1)

    if not hasattr(request.node, "stash"):  # compatibility with pytest<7.0.0
        request.node.stash = {}  # use regular dict instead of pytest.Stash
    request.node.stash[FIXTURE_OK] = True

    # Perform checks which may skip this test.
    check_net_interfaces()
    check_prerequisites()

    # Store the fact that this fixture hasn't successfully finished yet.
    # This is checked before temporary directory teardown to decide whether
    # it's okay to remove the directory.
    request.node.stash[FIXTURE_OK] = False

    setup_test()
    try:
        start_servers()
        mlogger.debug("executing test(s)")
        yield
    finally:
        mlogger.debug("test(s) finished")
        stop_servers()
        get_core_dumps()
        request.node.stash[FIXTURE_OK] = True



#!/usr/bin/python3

# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import mmap
import os
import subprocess
import sys
import time

import pytest

pytest.importorskip("dns", minversion="2.0.0")
import dns.exception
import dns.message
import dns.name
import dns.query
import dns.rcode
import dns.rdataclass
import dns.rdatatype
import dns.resolver


pytestmark = pytest.mark.skipif(
    sys.version_info < (3, 7), reason="Python >= 3.7 required [GL #3001]"
)


def has_signed_apex_nsec(zone, response):
    has_nsec = False
    has_rrsig = False

    ttl = 300
    nextname = "a."
    types = "NS SOA RRSIG NSEC DNSKEY"
    match = "{0} {1} IN NSEC {2}{0} {3}".format(zone, ttl, nextname, types)
    sig = "{0} {1} IN RRSIG NSEC 13 2 300".format(zone, ttl)

    for rr in response.answer:
        if match in rr.to_text():
            has_nsec = True
        if sig in rr.to_text():
            has_rrsig = True

    if not has_nsec:
        print("error: missing apex NSEC record in response")
    if not has_rrsig:
        print("error: missing NSEC signature in response")

    return has_nsec and has_rrsig


def do_query(server, qname, qtype, tcp=False):
    query = dns.message.make_query(qname, qtype, use_edns=True, want_dnssec=True)
    try:
        if tcp:
            response = dns.query.tcp(
                query, server.nameservers[0], timeout=3, port=server.port
            )
        else:
            response = dns.query.udp(
                query, server.nameservers[0], timeout=3, port=server.port
            )
    except dns.exception.Timeout:
        print(
            "error: query timeout for query {} {} to {}".format(
                qname, qtype, server.nameservers[0]
            )
        )
        return None

    return response


def verify_zone(zone, transfer):
    verify = os.getenv("VERIFY")
    assert verify is not None

    filename = "{}out".format(zone)
    with open(filename, "w", encoding="utf-8") as file:
        for rr in transfer.answer:
            file.write(rr.to_text())
            file.write("\n")

    # dnssec-verify command with default arguments.
    verify_cmd = [verify, "-z", "-o", zone, filename]

    verifier = subprocess.run(verify_cmd, capture_output=True, check=True)

    if verifier.returncode != 0:
        print("error: dnssec-verify {} failed".format(zone))
        sys.stderr.buffer.write(verifier.stderr)

    return verifier.returncode == 0


def read_statefile(server, zone):
    addr = server.nameservers[0]
    count = 0
    keyid = 0
    state = {}

    response = do_query(server, zone, "DS", tcp=True)
    if not isinstance(response, dns.message.Message):
        print("error: no response for {} DS from {}".format(zone, addr))
        return {}

    if response.rcode() == dns.rcode.NOERROR:
        # fetch key id from response.
        for rr in response.answer:
            if rr.match(
                dns.name.from_text(zone),
                dns.rdataclass.IN,
                dns.rdatatype.DS,
                dns.rdatatype.NONE,
            ):
                if count == 0:
                    keyid = list(dict(rr.items).items())[0][0].key_tag
                count += 1

        if count != 1:
            print(
                "error: expected a single DS in response for {} from {},"
                "got {}".format(zone, addr, count)
            )
            return {}
    else:
        print(
            "error: {} response for {} DNSKEY from {}".format(
                dns.rcode.to_text(response.rcode()), zone, addr
            )
        )
        return {}

    filename = "ns9/K{}+013+{:05d}.state".format(zone, keyid)
    print("read state file {}".format(filename))

    try:
        with open(filename, "r", encoding="utf-8") as file:
            for line in file:
                if line.startswith(";"):
                    continue
                key, val = line.strip().split(":", 1)
                state[key.strip()] = val.strip()

    except FileNotFoundError:
        # file may not be written just yet.
        return {}

    return state


def zone_check(server, zone):
    addr = server.nameservers[0]

    # wait until zone is fully signed.
    signed = False
    for _ in range(10):
        response = do_query(server, zone, "NSEC")
        if not isinstance(response, dns.message.Message):
            print("error: no response for {} NSEC from {}".format(zone, addr))
        elif response.rcode() == dns.rcode.NOERROR:
            signed = has_signed_apex_nsec(zone, response)
        else:
            print(
                "error: {} response for {} NSEC from {}".format(
                    dns.rcode.to_text(response.rcode()), zone, addr
                )
            )

        if signed:
            break

        time.sleep(1)

    assert signed

    # check if zone if DNSSEC valid.
    verified = False
    transfer = do_query(server, zone, "AXFR", tcp=True)
    if not isinstance(transfer, dns.message.Message):
        print("error: no response for {} AXFR from {}".format(zone, addr))
    elif transfer.rcode() == dns.rcode.NOERROR:
        verified = verify_zone(zone, transfer)
    else:
        print(
            "error: {} response for {} AXFR from {}".format(
                dns.rcode.to_text(transfer.rcode()), zone, addr
            )
        )

    assert verified


def keystate_check(server, zone, key):
    val = 0
    deny = False

    search = key
    if key.startswith("!"):
        deny = True
        search = key[1:]

    for _ in range(10):
        state = read_statefile(server, zone)
        try:
            val = state[search]
        except KeyError:
            pass

        if not deny and val != 0:
            break
        if deny and val == 0:
            break

        time.sleep(1)

    if deny:
        assert val == 0
    else:
        assert val != 0


def wait_for_log(filename, log):
    found = False

    for _ in range(10):
        print("read log file {}".format(filename))

        try:
            with open(filename, "r", encoding="utf-8") as file:
                s = mmap.mmap(file.fileno(), 0, access=mmap.ACCESS_READ)
                if s.find(bytes(log, "ascii")) != -1:
                    found = True
        except FileNotFoundError:
            print("file not found {}".format(filename))

        if found:
            break

        print("sleep")
        time.sleep(1)

    assert found


def test_checkds_dspublished(named_port):
    # We create resolver instances that will be used to send queries.
    server = dns.resolver.Resolver()
    server.nameservers = ["10.53.0.9"]
    server.port = named_port

    parent = dns.resolver.Resolver()
    parent.nameservers = ["10.53.0.2"]
    parent.port = named_port

    # DS correctly published in parent.
    zone_check(server, "dspublished.checkds.")
    wait_for_log(
        "ns9/named.run",
        "zone dspublished.checkds/IN (signed): checkds: DS response from 10.53.0.2",
    )
    keystate_check(parent, "dspublished.checkds.", "DSPublish")

    # DS correctly published in parent (reference to parental-agent).
    zone_check(server, "reference.checkds.")
    wait_for_log(
        "ns9/named.run",
        "zone reference.checkds/IN (signed): checkds: DS response from 10.53.0.2",
    )
    keystate_check(parent, "reference.checkds.", "DSPublish")

    # DS not published in parent.
    zone_check(server, "missing-dspublished.checkds.")
    wait_for_log(
        "ns9/named.run",
        "zone missing-dspublished.checkds/IN (signed): checkds: "
        "empty DS response from 10.53.0.5",
    )
    keystate_check(parent, "missing-dspublished.checkds.", "!DSPublish")

    # Badly configured parent.
    zone_check(server, "bad-dspublished.checkds.")
    wait_for_log(
        "ns9/named.run",
        "zone bad-dspublished.checkds/IN (signed): checkds: "
        "bad DS response from 10.53.0.6",
    )
    keystate_check(parent, "bad-dspublished.checkds.", "!DSPublish")

    # TBD: DS published in parent, but bogus signature.

    # DS correctly published in all parents.
    zone_check(server, "multiple-dspublished.checkds.")
    wait_for_log(
        "ns9/named.run",
        "zone multiple-dspublished.checkds/IN (signed): checkds: "
        "DS response from 10.53.0.2",
    )
    wait_for_log(
        "ns9/named.run",
        "zone multiple-dspublished.checkds/IN (signed): checkds: "
        "DS response from 10.53.0.4",
    )
    keystate_check(parent, "multiple-dspublished.checkds.", "DSPublish")

    # DS published in only one of multiple parents.
    zone_check(server, "incomplete-dspublished.checkds.")
    wait_for_log(
        "ns9/named.run",
        "zone incomplete-dspublished.checkds/IN (signed): checkds: "
        "DS response from 10.53.0.2",
    )
    wait_for_log(
        "ns9/named.run",
        "zone incomplete-dspublished.checkds/IN (signed): checkds: "
        "DS response from 10.53.0.4",
    )
    wait_for_log(
        "ns9/named.run",
        "zone incomplete-dspublished.checkds/IN (signed): checkds: "
        "empty DS response from 10.53.0.5",
    )
    keystate_check(parent, "incomplete-dspublished.checkds.", "!DSPublish")

    # One of the parents is badly configured.
    zone_check(server, "bad2-dswithdrawn.checkds.")
    wait_for_log(
        "ns9/named.run",
        "zone bad2-dspublished.checkds/IN (signed): checkds: "
        "DS response from 10.53.0.2",
    )
    wait_for_log(
        "ns9/named.run",
        "zone bad2-dspublished.checkds/IN (signed): checkds: "
        "DS response from 10.53.0.4",
    )
    wait_for_log(
        "ns9/named.run",
        "zone bad2-dspublished.checkds/IN (signed): checkds: "
        "bad DS response from 10.53.0.6",
    )
    keystate_check(parent, "bad2-dspublished.checkds.", "!DSPublish")

    # Check with resolver parental-agent.
    zone_check(server, "resolver-dspublished.checkds.")
    wait_for_log(
        "ns9/named.run",
        "zone resolver-dspublished.checkds/IN (signed): checkds: "
        "DS response from 10.53.0.3",
    )
    keystate_check(parent, "resolver-dspublished.checkds.", "DSPublish")

    # TBD: DS published in all parents, but one has bogus signature.

    # TBD: Check with TSIG

    # TBD: Check with TLS


def test_checkds_dswithdrawn(named_port):
    # We create resolver instances that will be used to send queries.
    server = dns.resolver.Resolver()
    server.nameservers = ["10.53.0.9"]
    server.port = named_port

    parent = dns.resolver.Resolver()
    parent.nameservers = ["10.53.0.2"]
    parent.port = named_port

    # DS correctly published in single parent.
    zone_check(server, "dswithdrawn.checkds.")
    wait_for_log(
        "ns9/named.run",
        "zone dswithdrawn.checkds/IN (signed): checkds: "
        "empty DS response from 10.53.0.5",
    )
    keystate_check(parent, "dswithdrawn.checkds.", "DSRemoved")

    # DS not withdrawn from parent.
    zone_check(server, "missing-dswithdrawn.checkds.")
    wait_for_log(
        "ns9/named.run",
        "zone missing-dswithdrawn.checkds/IN (signed): checkds: "
        "DS response from 10.53.0.2",
    )
    keystate_check(parent, "missing-dswithdrawn.checkds.", "!DSRemoved")

    # Badly configured parent.
    zone_check(server, "bad-dswithdrawn.checkds.")
    wait_for_log(
        "ns9/named.run",
        "zone bad-dswithdrawn.checkds/IN (signed): checkds: "
        "bad DS response from 10.53.0.6",
    )
    keystate_check(parent, "bad-dswithdrawn.checkds.", "!DSRemoved")

    # TBD: DS published in parent, but bogus signature.

    # DS correctly withdrawn from all parents.
    zone_check(server, "multiple-dswithdrawn.checkds.")
    wait_for_log(
        "ns9/named.run",
        "zone multiple-dswithdrawn.checkds/IN (signed): checkds: "
        "empty DS response from 10.53.0.5",
    )
    wait_for_log(
        "ns9/named.run",
        "zone multiple-dswithdrawn.checkds/IN (signed): checkds: "
        "empty DS response from 10.53.0.7",
    )
    keystate_check(parent, "multiple-dswithdrawn.checkds.", "DSRemoved")

    # DS withdrawn from only one of multiple parents.
    zone_check(server, "incomplete-dswithdrawn.checkds.")
    wait_for_log(
        "ns9/named.run",
        "zone incomplete-dswithdrawn.checkds/IN (signed): checkds: "
        "DS response from 10.53.0.2",
    )
    wait_for_log(
        "ns9/named.run",
        "zone incomplete-dswithdrawn.checkds/IN (signed): checkds: "
        "empty DS response from 10.53.0.5",
    )
    wait_for_log(
        "ns9/named.run",
        "zone incomplete-dswithdrawn.checkds/IN (signed): checkds: "
        "empty DS response from 10.53.0.7",
    )
    keystate_check(parent, "incomplete-dswithdrawn.checkds.", "!DSRemoved")

    # One of the parents is badly configured.
    zone_check(server, "bad2-dswithdrawn.checkds.")
    wait_for_log(
        "ns9/named.run",
        "zone bad2-dswithdrawn.checkds/IN (signed): checkds: "
        "empty DS response from 10.53.0.5",
    )
    wait_for_log(
        "ns9/named.run",
        "zone bad2-dswithdrawn.checkds/IN (signed): checkds: "
        "empty DS response from 10.53.0.7",
    )
    wait_for_log(
        "ns9/named.run",
        "zone bad2-dswithdrawn.checkds/IN (signed): checkds: "
        "bad DS response from 10.53.0.6",
    )
    keystate_check(parent, "bad2-dswithdrawn.checkds.", "!DSRemoved")

    # Check with resolver parental-agent.
    zone_check(server, "resolver-dswithdrawn.checkds.")
    wait_for_log(
        "ns9/named.run",
        "zone resolver-dswithdrawn.checkds/IN (signed): checkds: "
        "empty DS response from 10.53.0.8",
    )
    keystate_check(parent, "resolver-dswithdrawn.checkds.", "DSRemoved")

    # TBD: DS withdrawn from all parents, but one has bogus signature.



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

############################################################################
# ans.py: See README.anspy for details.
############################################################################

from __future__ import print_function
import os
import sys
import signal
import socket
import select
from datetime import datetime, timedelta
import functools

import dns, dns.message, dns.query
from dns.rdatatype import *
from dns.rdataclass import *
from dns.rcode import *
from dns.name import *

############################################################################
# set up the RRs to be returned in the next answer
#
# the message contains up to two pipe-separated ('|') fields.
#
# the first field of the message is a comma-separated list
# of actions indicating what to put into the answer set
# (e.g., a dname, a cname, another cname, etc)
#
# supported actions:
# - cname (cname from the current name to a new one in the same domain)
# - dname (dname to a new domain, plus a synthesized cname)
# - xname ("external" cname, to a new name in a new domain)
#
# example: xname, dname, cname represents a CNAME to an external
# domain which is then answered by a DNAME and synthesized
# CNAME pointing to yet another domain, which is then answered
# by a CNAME within the same domain, and finally an answer
# to the query. each RR in the answer set has a corresponding
# RRSIG. these signatures are not valid, but will exercise the
# response parser.
#
# the second field is a comma-separated list of which RRs in the
# answer set to include in the answer, in which order. if prepended
# with 's', the number indicates which signature to include.
#
# examples: for the answer set "cname, cname, cname", an rr set
# '1, s1, 2, s2, 3, s3, 4, s4' indicates that all four RRs should
# be included in the answer, with siagntures, in the original
# order, while 4, s4, 3, s3, 2, s2, 1, s1' indicates the order
# should be reversed, 's3, s3, s3, s3' indicates that the third
# RRSIG should be repeated four times and everything else should
# be omitted, and so on.
#
# if there is no second field (i.e., no pipe symbol appears in
# the line) , the default is to send all answers and signatures.
# if a pipe symbol exists but the second field is empty, then
# nothing is sent at all.
############################################################################
actions = []
rrs = []


def ctl_channel(msg):
    global actions, rrs

    msg = msg.splitlines().pop(0)
    print("received control message: %s" % msg)

    msg = msg.split(b"|")
    if len(msg) == 0:
        return

    actions = [x.strip() for x in msg[0].split(b",")]
    n = functools.reduce(
        lambda n, act: (n + (2 if act == b"dname" else 1)), [0] + actions
    )

    if len(msg) == 1:
        rrs = []
        for i in range(n):
            for b in [False, True]:
                rrs.append((i, b))
        return

    rlist = [x.strip() for x in msg[1].split(b",")]
    rrs = []
    for item in rlist:
        if item[0] == b"s"[0]:
            i = int(item[1:].strip()) - 1
            if i > n:
                print("invalid index %d" + (i + 1))
                continue
            rrs.append((int(item[1:]) - 1, True))
        else:
            i = int(item) - 1
            if i > n:
                print("invalid index %d" % (i + 1))
                continue
            rrs.append((i, False))


############################################################################
# Respond to a DNS query.
############################################################################
def create_response(msg):
    m = dns.message.from_wire(msg)
    qname = m.question[0].name.to_text()
    labels = qname.lower().split(".")
    wantsigs = True if m.ednsflags & dns.flags.DO else False

    # get qtype
    rrtype = m.question[0].rdtype
    typename = dns.rdatatype.to_text(rrtype)

    # for 'www.example.com.'...
    # - name is 'www'
    # - domain is 'example.com.'
    # - sld is 'example'
    # - tld is 'com.'
    name = labels.pop(0)
    domain = ".".join(labels)
    sld = labels.pop(0)
    tld = ".".join(labels)

    print("query: " + qname + "/" + typename)
    print("domain: " + domain)

    # default answers, depending on QTYPE.
    # currently only A, AAAA, TXT and NS are supported.
    ttl = 86400
    additionalA = "10.53.0.4"
    additionalAAAA = "fd92:7065:b8e:ffff::4"
    if typename == "A":
        final = "10.53.0.4"
    elif typename == "AAAA":
        final = "fd92:7065:b8e:ffff::4"
    elif typename == "TXT":
        final = "Some\ text\ here"
    elif typename == "NS":
        domain = qname
        final = "ns1.%s" % domain
    else:
        final = None

    # RRSIG rdata - won't validate but will exercise response parsing
    t = datetime.now()
    delta = timedelta(30)
    t1 = t - delta
    t2 = t + delta
    inception = t1.strftime("%Y%m%d000000")
    expiry = t2.strftime("%Y%m%d000000")
    sigdata = "OCXH2De0yE4NMTl9UykvOsJ4IBGs/ZIpff2rpaVJrVG7jQfmj50otBAp A0Zo7dpBU4ofv0N/F2Ar6LznCncIojkWptEJIAKA5tHegf/jY39arEpO cevbGp6DKxFhlkLXNcw7k9o7DSw14OaRmgAjXdTFbrl4AiAa0zAttFko Tso="

    # construct answer set.
    answers = []
    sigs = []
    curdom = domain
    curname = name
    i = 0

    for action in actions:
        if name != "test":
            continue
        if action == b"xname":
            owner = curname + "." + curdom
            newname = "cname%d" % i
            i += 1
            newdom = "domain%d.%s" % (i, tld)
            i += 1
            target = newname + "." + newdom
            print("add external CNAME %s to %s" % (owner, target))
            answers.append(dns.rrset.from_text(owner, ttl, IN, CNAME, target))
            rrsig = "CNAME 5 3 %d %s %s 12345 %s %s" % (
                ttl,
                expiry,
                inception,
                domain,
                sigdata,
            )
            print("add external RRISG(CNAME) %s to %s" % (owner, target))
            sigs.append(dns.rrset.from_text(owner, ttl, IN, RRSIG, rrsig))
            curname = newname
            curdom = newdom
            continue

        if action == b"cname":
            owner = curname + "." + curdom
            newname = "cname%d" % i
            target = newname + "." + curdom
            i += 1
            print("add CNAME %s to %s" % (owner, target))
            answers.append(dns.rrset.from_text(owner, ttl, IN, CNAME, target))
            rrsig = "CNAME 5 3 %d %s %s 12345 %s %s" % (
                ttl,
                expiry,
                inception,
                domain,
                sigdata,
            )
            print("add RRSIG(CNAME) %s to %s" % (owner, target))
            sigs.append(dns.rrset.from_text(owner, ttl, IN, RRSIG, rrsig))
            curname = newname
            continue

        if action == b"dname":
            owner = curdom
            newdom = "domain%d.%s" % (i, tld)
            i += 1
            print("add DNAME %s to %s" % (owner, newdom))
            answers.append(dns.rrset.from_text(owner, ttl, IN, DNAME, newdom))
            rrsig = "DNAME 5 3 %d %s %s 12345 %s %s" % (
                ttl,
                expiry,
                inception,
                domain,
                sigdata,
            )
            print("add RRSIG(DNAME) %s to %s" % (owner, newdom))
            sigs.append(dns.rrset.from_text(owner, ttl, IN, RRSIG, rrsig))
            owner = curname + "." + curdom
            target = curname + "." + newdom
            print("add synthesized CNAME %s to %s" % (owner, target))
            answers.append(dns.rrset.from_text(owner, ttl, IN, CNAME, target))
            rrsig = "CNAME 5 3 %d %s %s 12345 %s %s" % (
                ttl,
                expiry,
                inception,
                domain,
                sigdata,
            )
            print("add synthesized RRSIG(CNAME) %s to %s" % (owner, target))
            sigs.append(dns.rrset.from_text(owner, ttl, IN, RRSIG, rrsig))
            curdom = newdom
            continue

    # now add the final answer
    owner = curname + "." + curdom
    answers.append(dns.rrset.from_text(owner, ttl, IN, rrtype, final))
    rrsig = "%s 5 3 %d %s %s 12345 %s %s" % (
        typename,
        ttl,
        expiry,
        inception,
        domain,
        sigdata,
    )
    sigs.append(dns.rrset.from_text(owner, ttl, IN, RRSIG, rrsig))

    # prepare the response and convert to wire format
    r = dns.message.make_response(m)

    if name != "test":
        r.answer.append(answers[-1])
        if wantsigs:
            r.answer.append(sigs[-1])
    else:
        for i, sig in rrs:
            if sig and not wantsigs:
                continue
            elif sig:
                r.answer.append(sigs[i])
            else:
                r.answer.append(answers[i])

    if typename != "NS":
        r.authority.append(
            dns.rrset.from_text(domain, ttl, IN, "NS", ("ns1.%s" % domain))
        )
    r.additional.append(
        dns.rrset.from_text(("ns1.%s" % domain), 86400, IN, A, additionalA)
    )
    r.additional.append(
        dns.rrset.from_text(("ns1.%s" % domain), 86400, IN, AAAA, additionalAAAA)
    )

    r.flags |= dns.flags.AA
    r.use_edns()
    return r.to_wire()


def sigterm(signum, frame):
    print("Shutting down now...")
    os.remove("ans.pid")
    running = False
    sys.exit(0)


############################################################################
# Main
#
# Set up responder and control channel, open the pid file, and start
# the main loop, listening for queries on the query channel or commands
# on the control channel and acting on them.
############################################################################
ip4 = "10.53.0.4"
ip6 = "fd92:7065:b8e:ffff::4"

try:
    port = int(os.environ["PORT"])
except:
    port = 5300

try:
    ctrlport = int(os.environ["EXTRAPORT1"])
except:
    ctrlport = 5300

query4_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
query4_socket.bind((ip4, port))

havev6 = True
try:
    query6_socket = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
    try:
        query6_socket.bind((ip6, port))
    except:
        query6_socket.close()
        havev6 = False
except:
    havev6 = False

ctrl_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
ctrl_socket.bind((ip4, ctrlport))
ctrl_socket.listen(5)

signal.signal(signal.SIGTERM, sigterm)

f = open("ans.pid", "w")
pid = os.getpid()
print(pid, file=f)
f.close()

running = True

print("Listening on %s port %d" % (ip4, port))
if havev6:
    print("Listening on %s port %d" % (ip6, port))
print("Control channel on %s port %d" % (ip4, ctrlport))
print("Ctrl-c to quit")

if havev6:
    input = [query4_socket, query6_socket, ctrl_socket]
else:
    input = [query4_socket, ctrl_socket]

while running:
    try:
        inputready, outputready, exceptready = select.select(input, [], [])
    except select.error as e:
        break
    except socket.error as e:
        break
    except KeyboardInterrupt:
        break

    for s in inputready:
        if s == ctrl_socket:
            # Handle control channel input
            conn, addr = s.accept()
            print("Control channel connected")
            while True:
                msg = conn.recv(65535)
                if not msg:
                    break
                ctl_channel(msg)
            conn.close()
        if s == query4_socket or s == query6_socket:
            print("Query received on %s" % (ip4 if s == query4_socket else ip6))
            # Handle incoming queries
            msg = s.recvfrom(65535)
            rsp = create_response(msg[0])
            if rsp:
                s.sendto(rsp, msg[1])
    if not running:
        break



############################################################################
# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.
############################################################################

"""
Utility to check ISC config grammar consistency. It detects statement names
which use different grammar depending on position in the configuration file.
E.g. "max-zone-ttl" in dnssec-policy uses '<duration>'
vs. '( unlimited | <duration> ) used in options.
"""

from collections import namedtuple
from itertools import groupby
import fileinput

import parsegrammar


def statement2block(grammar, path):
    """Return mapping statement name to "path" where it is allowed.
    _top is placeholder name for the namesless topmost context.

    E.g. {
        'options: [('_top',)],
        'server': [('_top', 'view'), ('_top',)],
        'rate-limit': [('_top', 'options'), ('_top', 'view')],
        'slip': [('_top', 'options', 'rate-limit'), ('_top', 'view', 'rate-limit')]
    }
    """
    key2place = {}

    for key in grammar:
        assert not key.startswith("_")
        key2place.setdefault(key, []).append(tuple(path))
        if "_mapbody" in grammar[key]:
            nested2block = statement2block(grammar[key]["_mapbody"], path + [key])
            # merge to uppermost output dictionary
            for nested_key, nested_path in nested2block.items():
                key2place.setdefault(nested_key, []).extend(nested_path)
    return key2place


def get_statement_grammar(grammar, path, name):
    """Descend into grammar dict using provided path
    and return final dict found there.

    Intermediate steps into "_mapbody" subkeys are done automatically.
    """
    assert path[0] == "_top"
    path = list(path) + [name]
    for step in path[1:]:
        if "_mapbody" in grammar:
            grammar = grammar["_mapbody"]
        grammar = grammar[step]
    return grammar


Statement = namedtuple("Statement", ["path", "name", "subgrammar"])


def groupby_grammar(statements):
    """
    Return groups of Statement tuples with identical grammars and flags.
    See itertools.groupby.
    """

    def keyfunc(statement):
        return sorted(statement.subgrammar.items())

    groups = []
    statements = sorted(statements, key=keyfunc)
    for _key, group in groupby(statements, keyfunc):
        groups.append(list(group))  # Store group iterator as a list
    return groups


def diff_statements(whole_grammar, places):
    """
    Return map {statement name: [groups of [Statement]s with identical grammar].
    """
    out = {}
    for statement_name, paths in places.items():
        grammars = []
        for path in paths:
            statement_grammar = get_statement_grammar(
                whole_grammar, path, statement_name
            )
            grammars.append(Statement(path, statement_name, statement_grammar))
        groups = groupby_grammar(grammars)
        out[statement_name] = groups
    return out


def pformat_grammar(node, level=1):
    """Pretty print a given grammar node in the same way as cfg_test would"""

    def sortkey(item):
        """Treat 'type' specially and always put it first, for zone types"""
        key, _ = item
        if key == "type":
            return ""
        return key

    if "_grammar" in node:  # no nesting
        assert "_id" not in node
        assert "_mapbody" not in node
        out = node["_grammar"] + ";"
        if "_flags" in node:
            out += " // " + ", ".join(node["_flags"])
        return out + "\n"

    # a nested map
    out = ""
    indent = level * "\t"
    if not node.get("_ignore_this_level"):
        if "_id" in node:
            out += node["_id"] + " "
        out += "{\n"

    for key, subnode in sorted(node["_mapbody"].items(), key=sortkey):
        if not subnode.get("_ignore_this_level"):
            out += f"{indent}{subnode.get('_pprint_name', key)}"
            inner_grammar = pformat_grammar(node["_mapbody"][key], level=level + 1)
        else:  # act as if we were not in a map
            inner_grammar = pformat_grammar(node["_mapbody"][key], level=level)
        if inner_grammar[0] != ";":  # we _did_ find some arguments
            out += " "
        out += inner_grammar

    if not node.get("_ignore_this_level"):
        out += indent[:-1] + "};"  # unindent the closing bracket
        if "_flags" in node:
            out += " // " + ", ".join(node["_flags"])
    return out + "\n"


def main():
    """
    Ingest output from cfg_test --grammar and print out statements which use
    different grammar in different contexts.
    """
    with fileinput.input() as filein:
        grammar = parsegrammar.parse_mapbody(filein)
    places = statement2block(grammar, ["_top"])

    for statementname, groups in diff_statements(grammar, places).items():
        if len(groups) > 1:
            print(f'statement "{statementname}" is inconsistent across blocks')
            for group in groups:
                print(
                    "- path:", ", ".join(" -> ".join(variant.path) for variant in group)
                )
                print(" ", pformat_grammar(group[0].subgrammar, level=1))
            print()


if __name__ == "__main__":
    main()



#!/usr/bin/python3

# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from concurrent.futures import ThreadPoolExecutor, as_completed
import os
import random
import signal
import subprocess
from string import ascii_lowercase as letters
import time

import pytest

pytest.importorskip("dns", minversion="2.0.0")
import dns.exception
import dns.resolver


def do_work(named_proc, resolver, rndc_cmd, kill_method, n_workers, n_queries):
    """Creates a number of A queries to run in parallel
    in order simulate a slightly more realistic test scenario.

    The main idea of this function is to create and send a bunch
    of A queries to a target named instance and during this process
    a request for shutting down named will be issued.

    In the process of shutting down named, a couple control connections
    are created (by launching rndc) to ensure that the crash was fixed.

    if kill_method=="rndc" named will be asked to shutdown by
    means of rndc stop.
    if kill_method=="sigterm" named will be killed by SIGTERM on
    POSIX systems or by TerminateProcess() on Windows systems.

    :param named_proc: named process instance
    :type named_proc: subprocess.Popen

    :param resolver: target resolver
    :type resolver: dns.resolver.Resolver

    :param rndc_cmd: rndc command with default arguments
    :type rndc_cmd: list of strings, e.g. ["rndc", "-p", "23750"]

    :kill_method: "rndc" or "sigterm"
    :type kill_method: str

    :param n_workers: Number of worker threads to create
    :type n_workers: int

    :param n_queries: Total number of queries to send
    :type n_queries: int
    """
    # pylint: disable-msg=too-many-arguments
    # pylint: disable-msg=too-many-locals

    # helper function, args must be a list or tuple with arguments to rndc.
    def launch_rndc(args):
        return subprocess.call(rndc_cmd + args, timeout=10)

    # We're going to execute queries in parallel by means of a thread pool.
    # dnspython functions block, so we need to circunvent that.
    with ThreadPoolExecutor(n_workers + 1) as executor:
        # Helper dict, where keys=Future objects and values are tags used
        # to process results later.
        futures = {}

        # 50% of work will be A queries.
        # 1 work will be rndc stop.
        # Remaining work will be rndc status (so we test parallel control
        #  connections that were crashing named).
        shutdown = True
        for i in range(n_queries):
            if i < (n_queries // 2):
                # Half work will be standard A queries.
                # Among those we split 50% queries relname='www',
                # 50% queries relname=random characters
                if random.randrange(2) == 1:
                    tag = "good"
                    relname = "www"
                else:
                    tag = "bad"
                    length = random.randint(4, 10)
                    relname = "".join(
                        letters[random.randrange(len(letters))] for i in range(length)
                    )

                qname = relname + ".test"
                futures[executor.submit(resolver.resolve, qname, "A")] = tag
            elif shutdown:  # We attempt to stop named in the middle
                shutdown = False
                if kill_method == "rndc":
                    futures[executor.submit(launch_rndc, ["stop"])] = "stop"
                else:
                    futures[executor.submit(named_proc.terminate)] = "kill"
            else:
                # We attempt to send couple rndc commands while named is
                # being shutdown
                futures[executor.submit(launch_rndc, ["status"])] = "status"

        ret_code = -1
        for future in as_completed(futures):
            try:
                result = future.result()
                # If tag is "stop", result is an instance of
                # subprocess.CompletedProcess, then we check returncode
                # attribute to know if rncd stop command finished successfully.
                #
                # if tag is "kill" then the main function will check if
                # named process exited gracefully after SIGTERM signal.
                if futures[future] == "stop":
                    ret_code = result

            except (
                dns.resolver.NXDOMAIN,
                dns.resolver.NoNameservers,
                dns.exception.Timeout,
            ):
                pass

        if kill_method == "rndc":
            assert ret_code == 0


def wait_for_named_loaded(resolver, retries=10):
    for _ in range(retries):
        try:
            resolver.resolve("version.bind", "TXT", "CH")
            return True
        except (dns.resolver.NoNameservers, dns.exception.Timeout):
            time.sleep(1)
    return False


def wait_for_proc_termination(proc, max_timeout=10):
    for _ in range(max_timeout):
        if proc.poll() is not None:
            return True
        time.sleep(1)

    proc.send_signal(signal.SIGABRT)
    for _ in range(max_timeout):
        if proc.poll() is not None:
            return True
        time.sleep(1)

    return False


# We test named shutting down using two methods:
# Method 1: using rndc ctop
# Method 2: killing with SIGTERM
# In both methods named should exit gracefully.
@pytest.mark.parametrize("kill_method", ["rndc", "sigterm"])
def test_named_shutdown(named_port, control_port, kill_method):
    # pylint: disable-msg=too-many-locals
    cfg_dir = os.path.join(os.getcwd(), "resolver")
    assert os.path.isdir(cfg_dir)

    cfg_file = os.path.join(cfg_dir, "named.conf")
    assert os.path.isfile(cfg_file)

    named = os.getenv("NAMED")
    assert named is not None

    rndc = os.getenv("RNDC")
    assert rndc is not None

    # rndc configuration resides in ../_common/rndc.conf
    rndc_cfg = os.path.join("..", "_common", "rndc.conf")
    assert os.path.isfile(rndc_cfg)

    # rndc command with default arguments.
    rndc_cmd = [rndc, "-c", rndc_cfg, "-p", str(control_port), "-s", "10.53.0.3"]

    # We create a resolver instance that will be used to send queries.
    resolver = dns.resolver.Resolver()
    resolver.nameservers = ["10.53.0.3"]
    resolver.port = named_port

    named_cmdline = [named, "-c", cfg_file, "-f"]
    with subprocess.Popen(named_cmdline, cwd=cfg_dir) as named_proc:
        try:
            assert named_proc.poll() is None, "named isn't running"
            assert wait_for_named_loaded(resolver)
            do_work(
                named_proc,
                resolver,
                rndc_cmd,
                kill_method,
                n_workers=12,
                n_queries=16,
            )
            assert wait_for_proc_termination(named_proc)
            assert named_proc.returncode == 0, "named crashed"
        finally:  # Ensure named is terminated in case of an exception
            named_proc.kill()



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from typing import Any

import dns.rcode
import dns.message


# compatiblity with dnspython<2.0.0
try:
    # In dnspython>=2.0.0, dns.rcode.Rcode class is available
    # pylint: disable=invalid-name
    dns_rcode = dns.rcode.Rcode  # type: Any
except AttributeError:
    # In dnspython<2.0.0, selected rcodes are available as integers directly
    # from dns.rcode
    dns_rcode = dns.rcode


def rcode(message: dns.message.Message, expected_rcode) -> None:
    assert message.rcode() == expected_rcode, str(message)


def noerror(message: dns.message.Message) -> None:
    rcode(message, dns_rcode.NOERROR)



#!/usr/bin/python3

# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

# pylint: disable=unused-variable

import socket
import struct
import time

import pytest

pytest.importorskip("dns", minversion="2.0.0")
import dns.message
import dns.query


TIMEOUT = 10


def create_msg(qname, qtype, edns=-1):
    msg = dns.message.make_query(qname, qtype, use_edns=edns)
    return msg


def timeout():
    return time.time() + TIMEOUT


def create_socket(host, port):
    sock = socket.create_connection((host, port), timeout=10)
    sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, True)
    return sock


def test_tcp_garbage(named_port):
    with create_socket("10.53.0.7", named_port) as sock:
        msg = create_msg("a.example.", "A")
        (sbytes, stime) = dns.query.send_tcp(sock, msg, timeout())
        (response, rtime) = dns.query.receive_tcp(sock, timeout())

        wire = msg.to_wire()
        assert len(wire) > 0

        # Send DNS message shorter than DNS message header (12),
        # this should cause the connection to be terminated
        sock.send(struct.pack("!H", 11))
        sock.send(struct.pack("!s", b"0123456789a"))

        with pytest.raises(EOFError):
            try:
                (sbytes, stime) = dns.query.send_tcp(sock, msg, timeout())
                (response, rtime) = dns.query.receive_tcp(sock, timeout())
            except ConnectionError as e:
                raise EOFError from e


def test_tcp_garbage_response(named_port):
    with create_socket("10.53.0.7", named_port) as sock:
        msg = create_msg("a.example.", "A")
        (sbytes, stime) = dns.query.send_tcp(sock, msg, timeout())
        (response, rtime) = dns.query.receive_tcp(sock, timeout())

        wire = msg.to_wire()
        assert len(wire) > 0

        # Send DNS response instead of DNS query, this should cause
        # the connection to be terminated

        rmsg = dns.message.make_response(msg)
        (sbytes, stime) = dns.query.send_tcp(sock, rmsg, timeout())

        with pytest.raises(EOFError):
            try:
                (response, rtime) = dns.query.receive_tcp(sock, timeout())
            except ConnectionError as e:
                raise EOFError from e


# Regression test for CVE-2022-0396
def test_close_wait(named_port):
    with create_socket("10.53.0.7", named_port) as sock:
        msg = create_msg("a.example.", "A")
        (sbytes, stime) = dns.query.send_tcp(sock, msg, timeout())
        (response, rtime) = dns.query.receive_tcp(sock, timeout())

        msg = dns.message.make_query("a.example.", "A", use_edns=0, payload=1232)
        (sbytes, stime) = dns.query.send_tcp(sock, msg, timeout())

        # Shutdown the socket, but ignore the other side closing the socket
        # first because we sent DNS message with EDNS0
        try:
            sock.shutdown(socket.SHUT_RDWR)
        except ConnectionError:
            pass
        except OSError:
            pass

    # BIND allows one TCP client, the part above sends DNS messaage with EDNS0
    # after the first query. BIND should react adequately because of
    # ns7/named.dropedns and close the socket, making room for the next
    # request. If it gets stuck in CLOSE_WAIT state, there is no connection
    # available for the query below and it will time out.
    with create_socket("10.53.0.7", named_port) as sock:
        msg = create_msg("a.example.", "A")
        (sbytes, stime) = dns.query.send_tcp(sock, msg, timeout())
        (response, rtime) = dns.query.receive_tcp(sock, timeout())



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from datetime import datetime, timedelta
from collections import defaultdict
from time import sleep
import os

import dns.message
import dns.query
import dns.rcode

import isctest


# ISO datetime format without msec
fmt = "%Y-%m-%dT%H:%M:%SZ"

# The constants were taken from BIND 9 source code (lib/dns/zone.c)
max_refresh = timedelta(seconds=2419200)  # 4 weeks
max_expires = timedelta(seconds=14515200)  # 24 weeks
dayzero = datetime.utcfromtimestamp(0).replace(microsecond=0)

# Wait for the secondary zone files to appear to extract their mtime
max_secondary_zone_waittime_sec = 5


# Generic helper functions
def check_expires(expires, min_time, max_time):
    assert expires >= min_time
    assert expires <= max_time


def check_refresh(refresh, min_time, max_time):
    assert refresh >= min_time
    assert refresh <= max_time


def check_loaded(loaded, expected, now):
    # Sanity check the zone timers values
    assert (loaded - expected).total_seconds() < max_secondary_zone_waittime_sec
    assert loaded <= now


def check_zone_timers(loaded, expires, refresh, loaded_exp):
    now = datetime.utcnow().replace(microsecond=0)
    # Sanity checks the zone timers values
    if expires is not None:
        check_expires(expires, now, now + max_expires)
    if refresh is not None:
        check_refresh(refresh, now, now + max_refresh)
    check_loaded(loaded, loaded_exp, now)


#
# The output is gibberish, but at least make sure it does not crash.
#
def check_manykeys(name, zone=None):
    # pylint: disable=unused-argument
    assert name == "manykeys"


def zone_mtime(zonedir, name):
    try:
        si = os.stat(os.path.join(zonedir, "{}.db".format(name)))
    except FileNotFoundError:
        return dayzero

    mtime = datetime.utcfromtimestamp(si.st_mtime).replace(microsecond=0)

    return mtime


def test_zone_timers_primary(fetch_zones, load_timers, **kwargs):
    statsip = kwargs["statsip"]
    statsport = kwargs["statsport"]
    zonedir = kwargs["zonedir"]

    zones = fetch_zones(statsip, statsport)

    for zone in zones:
        (name, loaded, expires, refresh) = load_timers(zone, True)
        mtime = zone_mtime(zonedir, name)
        check_zone_timers(loaded, expires, refresh, mtime)


def test_zone_timers_secondary(fetch_zones, load_timers, **kwargs):
    statsip = kwargs["statsip"]
    statsport = kwargs["statsport"]
    zonedir = kwargs["zonedir"]

    # If any one of the zone files isn't ready, then retry until timeout.
    tries = max_secondary_zone_waittime_sec
    while tries >= 0:
        zones = fetch_zones(statsip, statsport)
        again = False
        for zone in zones:
            (name, loaded, expires, refresh) = load_timers(zone, False)
            mtime = zone_mtime(zonedir, name)
            if (mtime != dayzero) or (tries == 0):
                # mtime was either retrieved successfully or no tries were
                # left, run the check anyway.
                check_zone_timers(loaded, expires, refresh, mtime)
            else:
                tries = tries - 1
                again = True
                break
        if again:
            sleep(1)
        else:
            break


def test_zone_with_many_keys(fetch_zones, load_zone, **kwargs):
    statsip = kwargs["statsip"]
    statsport = kwargs["statsport"]

    zones = fetch_zones(statsip, statsport)

    for zone in zones:
        name = load_zone(zone)
        if name == "manykeys":
            check_manykeys(name)


def create_msg(qname, qtype):
    msg = dns.message.make_query(
        qname, qtype, want_dnssec=True, use_edns=0, payload=4096
    )

    return msg


def create_expected(data):
    expected = {
        "dns-tcp-requests-sizes-received-ipv4": defaultdict(int),
        "dns-tcp-responses-sizes-sent-ipv4": defaultdict(int),
        "dns-tcp-requests-sizes-received-ipv6": defaultdict(int),
        "dns-tcp-responses-sizes-sent-ipv6": defaultdict(int),
        "dns-udp-requests-sizes-received-ipv4": defaultdict(int),
        "dns-udp-requests-sizes-received-ipv6": defaultdict(int),
        "dns-udp-responses-sizes-sent-ipv4": defaultdict(int),
        "dns-udp-responses-sizes-sent-ipv6": defaultdict(int),
    }

    for k, v in data.items():
        for kk, vv in v.items():
            expected[k][kk] += vv

    return expected


def update_expected(expected, key, msg):
    msg_len = len(msg.to_wire())
    bucket_num = (msg_len // 16) * 16
    bucket = "{}-{}".format(bucket_num, bucket_num + 15)

    expected[key][bucket] += 1


def check_traffic(data, expected):
    def ordered(obj):
        if isinstance(obj, dict):
            return sorted((k, ordered(v)) for k, v in obj.items())
        if isinstance(obj, list):
            return sorted(ordered(x) for x in obj)
        return obj

    ordered_data = ordered(data)
    ordered_expected = ordered(expected)

    assert len(ordered_data) == 8
    assert len(ordered_expected) == 8
    assert len(data) == len(ordered_data)
    assert len(expected) == len(ordered_expected)

    assert ordered_data == ordered_expected


def test_traffic(fetch_traffic, **kwargs):
    statsip = kwargs["statsip"]
    statsport = kwargs["statsport"]

    data = fetch_traffic(statsip, statsport)
    exp = create_expected(data)

    msg = create_msg("short.example.", "TXT")
    update_expected(exp, "dns-udp-requests-sizes-received-ipv4", msg)
    ans = isctest.query.udp(msg, statsip)
    isctest.check.noerror(ans)
    update_expected(exp, "dns-udp-responses-sizes-sent-ipv4", ans)
    data = fetch_traffic(statsip, statsport)

    check_traffic(data, exp)

    msg = create_msg("long.example.", "TXT")
    update_expected(exp, "dns-udp-requests-sizes-received-ipv4", msg)
    ans = isctest.query.udp(msg, statsip)
    isctest.check.noerror(ans)
    update_expected(exp, "dns-udp-responses-sizes-sent-ipv4", ans)
    data = fetch_traffic(statsip, statsport)

    check_traffic(data, exp)

    msg = create_msg("short.example.", "TXT")
    update_expected(exp, "dns-tcp-requests-sizes-received-ipv4", msg)
    ans = isctest.query.tcp(msg, statsip)
    isctest.check.noerror(ans)
    update_expected(exp, "dns-tcp-responses-sizes-sent-ipv4", ans)
    data = fetch_traffic(statsip, statsport)

    check_traffic(data, exp)

    msg = create_msg("long.example.", "TXT")
    update_expected(exp, "dns-tcp-requests-sizes-received-ipv4", msg)
    ans = isctest.query.tcp(msg, statsip)
    isctest.check.noerror(ans)
    update_expected(exp, "dns-tcp-responses-sizes-sent-ipv4", ans)
    data = fetch_traffic(statsip, statsport)

    check_traffic(data, exp)



#!/usr/bin/python3

# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import shutil
import subprocess

import pytest


@pytest.fixture
def gnutls_cli_executable():
    # Ensure gnutls-cli is available.
    executable = shutil.which("gnutls-cli")
    if not executable:
        pytest.skip("gnutls-cli not found in PATH")

    # Ensure gnutls-cli supports the --logfile command-line option.
    output = subprocess.run(
        [executable, "--logfile=/dev/null"],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        check=False,
    ).stdout
    if b"illegal option" in output:
        pytest.skip("gnutls-cli does not support the --logfile option")

    return executable


@pytest.fixture
def sslyze_executable():
    # Check whether sslyze is available.
    executable = shutil.which("sslyze")
    if not executable:
        pytest.skip("sslyze not found in PATH")

    return executable



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import glob
import struct


class RawFormatHeader(dict):
    """
    A dictionary of raw-format header fields read from a zone file.
    """

    fields = [
        "format",
        "version",
        "dumptime",
        "flags",
        "sourceserial",
        "lastxfrin",
    ]

    def __init__(self, file_name):
        header = struct.Struct(">IIIIII")
        with open(file_name, "rb") as data:
            header_data = data.read(header.size)
        super().__init__(zip(self.fields, header.unpack_from(header_data)))


def test_unsigned_serial_number():
    """
    Check whether all signed zone files in the "ns8" subdirectory contain the
    serial number of the unsigned version of the zone in the raw-format header.
    The test assumes that all "*.signed" files in the "ns8" subdirectory are in
    raw format.

    Notes:

      - The actual zone signing and dumping happens while the tests.sh phase of
        the "inline" system test is set up and run.  This check only verifies
        the outcome of those events; it does not initiate any signing or
        dumping itself.

      - example[0-9][0-9].com.db.signed files are initially signed by
        dnssec-signzone while the others - by named.
    """

    zones_with_unsigned_serial_missing = []

    for signed_zone in sorted(glob.glob("ns8/*.signed")):
        raw_header = RawFormatHeader(signed_zone)
        # Ensure the unsigned serial number is placed where it is expected.
        assert raw_header["format"] == 2
        assert raw_header["version"] == 1
        # Check whether the header flags indicate that the unsigned serial
        # number is set and that the latter is indeed set.
        if raw_header["flags"] & 0x02 == 0 or raw_header["sourceserial"] == 0:
            zones_with_unsigned_serial_missing.append(signed_zone)

    assert not zones_with_unsigned_serial_missing



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import sys

try:
    import yaml
# pylint: disable=bare-except
except:
    print("No python yaml module, skipping")
    sys.exit(1)

with open(sys.argv[1], "r", encoding="utf-8") as f:
    for item in yaml.safe_load_all(f):
        for key in sys.argv[2:]:
            try:
                key = int(key)
            except ValueError:
                pass

            try:
                item = item[key]
            except KeyError:
                print('Key "' + key + '" not found.')
                sys.exit(1)

        print(item)



############################################################################
# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.
############################################################################

"""
Sphinx domain "namedconf". See iscconf.py for details.

"""
from docutils import nodes

import iscconf
import mergegrammar


class ToBeReplacedStatementList(nodes.General, nodes.Element):
    """
    Placeholder, does nothing, but must be picklable
    (= cannot be in generated class).
    """


def setup(app):
    grammar = mergegrammar.combine()
    return iscconf.setup(
        app, "namedconf", "named.conf", ToBeReplacedStatementList, grammar
    )



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import sys

try:
    import yaml
except (ModuleNotFoundError, ImportError):
    print("No python yaml module, skipping")
    sys.exit(1)

import subprocess
import pprint

DNSTAP_READ = sys.argv[1]
DATAFILE = sys.argv[2]
ARGS = [DNSTAP_READ, "-y", DATAFILE]

with subprocess.Popen(ARGS, stdout=subprocess.PIPE) as f:
    for y in yaml.load_all(f.stdout, Loader=yaml.SafeLoader):
        pprint.pprint(y)



#!/usr/bin/python3

# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

# This script is a 'port' broker.  It keeps track of ports given to the
# individual system subtests, so every test is given a unique port range.

import logging
import os
from pathlib import Path
import platform
import random
import subprocess
import time
from typing import Dict, List, NamedTuple, Union

# Uncomment to enable DEBUG logging
# logging.basicConfig(
#     format="get_algorithms.py %(levelname)s %(message)s", level=logging.DEBUG
# )

STABLE_PERIOD = 3600 * 3
"""number of secs during which algorithm selection remains stable"""


class Algorithm(NamedTuple):
    name: str
    number: int
    bits: int


class AlgorithmSet(NamedTuple):
    """Collection of DEFAULT, ALTERNATIVE and DISABLED algorithms"""

    default: Union[Algorithm, List[Algorithm]]
    """DEFAULT is the algorithm for testing."""

    alternative: Union[Algorithm, List[Algorithm]]
    """ALTERNATIVE is an alternative algorithm for test cases that require more
    than one algorithm (for example algorithm rollover)."""

    disabled: Union[Algorithm, List[Algorithm]]
    """DISABLED is an algorithm that is used for tests against the
    "disable-algorithms" configuration option."""


RSASHA1 = Algorithm("RSASHA1", 5, 1280)
RSASHA256 = Algorithm("RSASHA256", 8, 1280)
RSASHA512 = Algorithm("RSASHA512", 10, 1280)
ECDSAP256SHA256 = Algorithm("ECDSAP256SHA256", 13, 256)
ECDSAP384SHA384 = Algorithm("ECDSAP384SHA384", 14, 384)
ED25519 = Algorithm("ED25519", 15, 256)
ED448 = Algorithm("ED448", 16, 456)

ALL_ALGORITHMS = [
    RSASHA1,
    RSASHA256,
    RSASHA512,
    ECDSAP256SHA256,
    ECDSAP384SHA384,
    ED25519,
    ED448,
]

ALGORITHM_SETS = {
    "stable": AlgorithmSet(
        default=ECDSAP256SHA256, alternative=RSASHA256, disabled=ECDSAP384SHA384
    ),
    "ecc_default": AlgorithmSet(
        default=[
            ECDSAP256SHA256,
            ECDSAP384SHA384,
            ED25519,
            ED448,
        ],
        alternative=RSASHA256,
        disabled=RSASHA512,
    ),
    # FUTURE The system tests needs more work before they're ready for this.
    # "random": AlgorithmSet(
    #     default=ALL_ALGORITHMS,
    #     alternative=ALL_ALGORITHMS,
    #     disabled=ALL_ALGORITHMS,
    # ),
}

TESTCRYPTO = Path(__file__).resolve().parent / "testcrypto.sh"

KEYGEN = os.getenv("KEYGEN", "")
if not KEYGEN:
    raise RuntimeError("KEYGEN environment variable has to be set")

ALGORITHM_SET = os.getenv("ALGORITHM_SET", "stable")
assert ALGORITHM_SET in ALGORITHM_SETS, f'ALGORITHM_SET "{ALGORITHM_SET}" unknown'
logging.debug('choosing from ALGORITHM_SET "%s"', ALGORITHM_SET)


def is_supported(alg: Algorithm) -> bool:
    """Test whether a given algorithm is supported on the current platform."""
    try:
        subprocess.run(
            f"{TESTCRYPTO} -q {alg.name}",
            shell=True,
            check=True,
            env={
                "KEYGEN": KEYGEN,
                "TMPDIR": os.getenv("TMPDIR", "/tmp"),
            },
            stdout=subprocess.DEVNULL,
        )
    except subprocess.CalledProcessError as exc:
        logging.debug(exc)
        logging.info("algorithm %s not supported", alg.name)
        return False
    return True


def filter_supported(algs: AlgorithmSet) -> AlgorithmSet:
    """Select supported algorithms from the set."""
    filtered = {}
    for alg_type in algs._fields:
        candidates = getattr(algs, alg_type)
        if isinstance(candidates, Algorithm):
            candidates = [candidates]
        supported = list(filter(is_supported, candidates))
        if len(supported) == 1:
            supported = supported.pop()
        elif not supported:
            raise RuntimeError(
                f'no {alg_type.upper()} algorithm from "{ALGORITHM_SET}" set '
                "supported on this platform"
            )
        filtered[alg_type] = supported
    return AlgorithmSet(**filtered)


def select_random(algs: AlgorithmSet, stable_period=STABLE_PERIOD) -> AlgorithmSet:
    """Select random DEFAULT, ALTERNATIVE and DISABLED algorithms from the set.

    The algorithm selection is deterministic for a given time period and
    platform. This should make potential issues more reproducible.

    To increase the likelyhood of detecting an issue with a given algorithm in
    CI, the current platform is used as a randomness source. When testing on
    multiple platforms at the same time, this ensures more algorithm variance
    while keeping reproducibility for a single platform.

    The function also ensures that DEFAULT, ALTERNATIVE and DISABLED algorithms
    are all different.
    """
    # FUTURE Random selection of ALTERNATIVE and DISABLED algorithms needs to
    # be implemented.
    alternative = algs.alternative
    disabled = algs.disabled
    assert isinstance(
        alternative, Algorithm
    ), "ALTERNATIVE algorithm randomization not supported yet"
    assert isinstance(
        disabled, Algorithm
    ), "DISABLED algorithm randomization not supported yet"

    # initialize randomness
    now = time.time()
    time_seed = int(now - now % stable_period)
    seed = f"{platform.platform()}_{time_seed}"
    random.seed(seed)

    # DEFAULT selection
    if isinstance(algs.default, Algorithm):
        default = algs.default
    else:
        candidates = algs.default
        for taken in [alternative, disabled]:
            try:
                candidates.remove(taken)
            except ValueError:
                pass
        assert len(candidates), "no possible choice for DEFAULT algorithm"
        random.shuffle(candidates)
        default = candidates[0]

    # Ensure only single algorithm is present for each option
    assert isinstance(default, Algorithm)
    assert isinstance(alternative, Algorithm)
    assert isinstance(disabled, Algorithm)

    assert default != alternative, "DEFAULT and ALTERNATIVE algorithms are the same"
    assert default != disabled, "DEFAULT and DISABLED algorithms are the same"
    assert alternative != disabled, "ALTERNATIVE and DISABLED algorithms are the same"

    return AlgorithmSet(default, alternative, disabled)


def algorithms_env(algs: AlgorithmSet) -> Dict[str, str]:
    """Return environment variables with selected algorithms as a dict."""
    algs_env: Dict[str, str] = {}

    def set_alg_env(alg: Algorithm, prefix):
        algs_env[f"{prefix}_ALGORITHM"] = alg.name
        algs_env[f"{prefix}_ALGORITHM_NUMBER"] = str(alg.number)
        algs_env[f"{prefix}_BITS"] = str(alg.bits)

    assert isinstance(algs.default, Algorithm)
    assert isinstance(algs.alternative, Algorithm)
    assert isinstance(algs.disabled, Algorithm)

    set_alg_env(algs.default, "DEFAULT")
    set_alg_env(algs.alternative, "ALTERNATIVE")
    set_alg_env(algs.disabled, "DISABLED")

    logging.info("selected algorithms: %s", algs_env)
    return algs_env


def main():
    try:
        algs = ALGORITHM_SETS[ALGORITHM_SET]
        algs = filter_supported(algs)
        algs = select_random(algs)
        algs_env = algorithms_env(algs)
    except Exception:
        # if anything goes wrong, the conf.sh ignores error codes, so make sure
        # we set an environment variable to an error value that can be checked
        # later by the test runner and/or tests themselves
        print("export ALGORITHM_SET=error")
        raise
    for name, value in algs_env.items():
        print(f"export {name}={value}")


if __name__ == "__main__":
    main()



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from __future__ import print_function
import os
import sys
import signal
import socket
import select
import struct

import dns, dns.message
from dns.rcode import *

modes = [
    b"silent",  # Do not respond
    b"close",  # UDP: same as silent; TCP: also close the connection
    b"servfail",  # Always respond with SERVFAIL
    b"unstable",  # Constantly switch between "silent" and "servfail"
]
mode = modes[0]
n = 0


def ctrl_channel(msg):
    global modes, mode, n

    msg = msg.splitlines().pop(0)
    print("Received control message: %s" % msg)

    if msg in modes:
        mode = msg
        n = 0
        print("New mode: %s" % str(mode))


def create_servfail(msg):
    m = dns.message.from_wire(msg)
    qname = m.question[0].name.to_text()
    rrtype = m.question[0].rdtype
    typename = dns.rdatatype.to_text(rrtype)

    with open("query.log", "a") as f:
        f.write("%s %s\n" % (typename, qname))
        print("%s %s" % (typename, qname), end=" ")

    r = dns.message.make_response(m)
    r.set_rcode(SERVFAIL)
    return r


def sigterm(signum, frame):
    print("Shutting down now...")
    os.remove("ans.pid")
    running = False
    sys.exit(0)


ip4 = "10.53.0.8"

try:
    port = int(os.environ["PORT"])
except:
    port = 5300

try:
    ctrlport = int(os.environ["EXTRAPORT1"])
except:
    ctrlport = 5300

query4_udp = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
query4_udp.bind((ip4, port))

query4_tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
query4_tcp.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
query4_tcp.bind((ip4, port))
query4_tcp.listen(100)

ctrl4_tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
ctrl4_tcp.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
ctrl4_tcp.bind((ip4, ctrlport))
ctrl4_tcp.listen(100)

signal.signal(signal.SIGTERM, sigterm)

f = open("ans.pid", "w")
pid = os.getpid()
print(pid, file=f)
f.close()

running = True

print("Listening on %s port %d" % (ip4, port))
print("Listening on %s port %d" % (ip4, ctrlport))
print("Ctrl-c to quit")

input = [query4_udp, query4_tcp, ctrl4_tcp]

hung_conns = []

while running:
    try:
        inputready, outputready, exceptready = select.select(input, [], [])
    except select.error as e:
        break
    except socket.error as e:
        break
    except KeyboardInterrupt:
        break

    for s in inputready:
        if s == query4_udp:
            n = n + 1
            print("UDP query received on %s" % ip4, end=" ")
            msg = s.recvfrom(65535)
            if (
                mode == b"silent"
                or mode == b"close"
                or (mode == b"unstable" and n % 2 == 1)
            ):
                # Do not respond.
                print("NO RESPONSE (%s)" % str(mode))
                continue
            elif mode == b"servfail" or (mode == b"unstable" and n % 2 == 0):
                rsp = create_servfail(msg[0])
                if rsp:
                    print(dns.rcode.to_text(rsp.rcode()))
                    s.sendto(rsp.to_wire(), msg[1])
                else:
                    print("NO RESPONSE (can not create a response)")
            else:
                raise (Exception("unsupported mode: %s" % mode))
        elif s == query4_tcp:
            n = n + 1
            print("TCP query received on %s" % ip4, end=" ")
            conn = None
            try:
                if mode == b"silent" or (mode == b"unstable" and n % 2 == 1):
                    conn, addr = s.accept()
                    # Do not respond and hang the connection.
                    print("NO RESPONSE (%s)" % str(mode))
                    hung_conns.append(conn)
                    continue
                elif mode == b"close":
                    conn, addr = s.accept()
                    # Do not respond and close the connection.
                    print("NO RESPONSE (%s)" % str(mode))
                    conn.close()
                    continue
                elif mode == b"servfail" or (mode == b"unstable" and n % 2 == 0):
                    conn, addr = s.accept()
                    # get TCP message length
                    msg = conn.recv(2)
                    if len(msg) != 2:
                        print("NO RESPONSE (can not read the message length)")
                        conn.close()
                        continue
                    length = struct.unpack(">H", msg[:2])[0]
                    msg = conn.recv(length)
                    if len(msg) != length:
                        print("NO RESPONSE (can not read the message)")
                        conn.close()
                        continue
                    rsp = create_servfail(msg)
                    if rsp:
                        print(dns.rcode.to_text(rsp.rcode()))
                        wire = rsp.to_wire()
                        conn.send(struct.pack(">H", len(wire)))
                        conn.send(wire)
                    else:
                        print("NO RESPONSE (can not create a response)")
                else:
                    raise (Exception("unsupported mode: %s" % mode))
            except socket.error as e:
                print("NO RESPONSE (error: %s)" % str(e))
            if conn:
                conn.close()
        elif s == ctrl4_tcp:
            print("Control channel connected")
            conn = None
            try:
                # Handle control channel input
                conn, addr = s.accept()
                msg = conn.recv(1024)
                if msg:
                    ctrl_channel(msg)
                conn.close()
            except s.timeout:
                pass
            if conn:
                conn.close()

    if not running:
        break



#!/usr/bin/python3

# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import time
import os

import pytest

pytest.importorskip("dns", minversion="2.0.0")
import dns.resolver


def wait_for_transfer(ip, port, client_ip, name, rrtype):
    resolver = dns.resolver.Resolver()
    resolver.nameservers = [ip]
    resolver.port = port

    for _ in range(10):
        try:
            resolver.resolve(name, rrtype, source=client_ip)
        except dns.resolver.NoNameservers:
            time.sleep(1)
        else:
            break
    else:
        raise RuntimeError(
            "zone transfer failed: "
            f"client {client_ip} got NXDOMAIN for {name} {rrtype} from @{ip}:{port}"
        )


def test_rpz_multiple_views(named_port):
    resolver = dns.resolver.Resolver()
    resolver.nameservers = ["10.53.0.3"]
    resolver.port = named_port

    wait_for_transfer("10.53.0.3", named_port, "10.53.0.2", "rpz-external.local", "SOA")
    wait_for_transfer("10.53.0.3", named_port, "10.53.0.5", "rpz-external.local", "SOA")

    # For 10.53.0.1 source IP:
    # - baddomain.com isn't allowed (CNAME .), should return NXDOMAIN
    # - gooddomain.com is allowed
    # - allowed. is allowed
    with pytest.raises(dns.resolver.NXDOMAIN):
        resolver.resolve("baddomain.", "A", source="10.53.0.1")

    ans = resolver.resolve("gooddomain.", "A", source="10.53.0.1")
    assert ans[0].address == "10.53.0.2"

    ans = resolver.resolve("allowed.", "A", source="10.53.0.1")
    assert ans[0].address == "10.53.0.2"

    # For 10.53.0.2 source IP:
    # - allowed.com isn't allowed (CNAME .), should return NXDOMAIN
    # - baddomain.com is allowed
    # - gooddomain.com is allowed
    ans = resolver.resolve("baddomain.", "A", source="10.53.0.2")
    assert ans[0].address == "10.53.0.2"

    ans = resolver.resolve("gooddomain.", "A", source="10.53.0.2")
    assert ans[0].address == "10.53.0.2"

    with pytest.raises(dns.resolver.NXDOMAIN):
        resolver.resolve("allowed.", "A", source="10.53.0.2")

    # For 10.53.0.3 source IP:
    # - gooddomain.com is allowed
    # - baddomain.com is allowed
    # - allowed. is allowed
    ans = resolver.resolve("baddomain.", "A", source="10.53.0.3")
    assert ans[0].address == "10.53.0.2"

    ans = resolver.resolve("gooddomain.", "A", source="10.53.0.3")
    assert ans[0].address == "10.53.0.2"

    ans = resolver.resolve("allowed.", "A", source="10.53.0.3")
    assert ans[0].address == "10.53.0.2"

    # For 10.53.0.4 source IP:
    # - gooddomain.com isn't allowed (CNAME .), should return NXDOMAIN
    # - baddomain.com isn't allowed (CNAME .), should return NXDOMAIN
    # - allowed. is allowed
    with pytest.raises(dns.resolver.NXDOMAIN):
        resolver.resolve("baddomain.", "A", source="10.53.0.4")

    with pytest.raises(dns.resolver.NXDOMAIN):
        resolver.resolve("gooddomain.", "A", source="10.53.0.4")

    ans = resolver.resolve("allowed.", "A", source="10.53.0.4")
    assert ans[0].address == "10.53.0.2"

    # For 10.53.0.5 (any) source IP:
    # - baddomain.com is allowed
    # - gooddomain.com isn't allowed (CNAME .), should return NXDOMAIN
    # - allowed.com isn't allowed (CNAME .), should return NXDOMAIN
    ans = resolver.resolve("baddomain.", "A", source="10.53.0.5")
    assert ans[0].address == "10.53.0.2"

    with pytest.raises(dns.resolver.NXDOMAIN):
        resolver.resolve("gooddomain.", "A", source="10.53.0.5")

    with pytest.raises(dns.resolver.NXDOMAIN):
        resolver.resolve("allowed.", "A", source="10.53.0.5")


def test_rpz_passthru_logging(named_port):
    resolver = dns.resolver.Resolver()
    resolver.nameservers = ["10.53.0.3"]
    resolver.port = named_port

    # Should generate a log entry into rpz_passthru.txt
    ans = resolver.resolve("allowed.", "A", source="10.53.0.1")
    assert ans[0].address == "10.53.0.2"

    # baddomain.com isn't allowed (CNAME .), should return NXDOMAIN
    # Should generate a log entry into rpz.txt
    with pytest.raises(dns.resolver.NXDOMAIN):
        resolver.resolve("baddomain.", "A", source="10.53.0.1")

    rpz_passthru_logfile = os.path.join("ns3", "rpz_passthru.txt")
    rpz_logfile = os.path.join("ns3", "rpz.txt")

    assert os.path.isfile(rpz_passthru_logfile)
    assert os.path.isfile(rpz_logfile)

    with open(rpz_passthru_logfile, encoding="utf-8") as log_file:
        line = log_file.read()
        assert "rpz QNAME PASSTHRU rewrite allowed/A/IN" in line

    with open(rpz_logfile, encoding="utf-8") as log_file:
        line = log_file.read()
        assert "rpz QNAME PASSTHRU rewrite allowed/A/IN" not in line
        assert "rpz QNAME NXDOMAIN rewrite baddomain/A/IN" in line



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import concurrent.futures
import os
import subprocess
import time


def run_rndc(server, rndc_command):
    """
    Send the specified 'rndc_command' to 'server' with a timeout of 10 seconds
    """
    rndc = os.getenv("RNDC")
    port = os.getenv("CONTROLPORT")

    cmdline = [rndc, "-c", "../_common/rndc.conf", "-p", port, "-s", server]
    cmdline.extend(rndc_command)

    subprocess.check_output(cmdline, stderr=subprocess.STDOUT, timeout=10)


def rndc_loop(test_state, domain):
    """
    Run "rndc addzone", "rndc modzone", and "rndc delzone" in a tight loop
    until the test is considered finished, ignoring errors
    """
    rndc_commands = [
        ["addzone", domain, '{ type primary; file "example.db"; };'],
        [
            "modzone",
            domain,
            '{ type primary; file "example.db"; allow-transfer { any; }; };',
        ],
        ["delzone", domain],
    ]

    while not test_state["finished"]:
        for command in rndc_commands:
            try:
                run_rndc("10.53.0.3", command)
            except subprocess.SubprocessError:
                pass


def check_if_server_is_responsive():
    """
    Check if server status can be successfully retrieved using "rndc status"
    """
    try:
        run_rndc("10.53.0.3", ["status"])
        return True
    except subprocess.SubprocessError:
        return False


def test_rndc_deadlock():
    """
    Test whether running "rndc addzone", "rndc modzone", and "rndc delzone"
    commands concurrently does not trigger a deadlock
    """
    test_state = {"finished": False}

    # Create 4 worker threads running "rndc" commands in a loop.
    with concurrent.futures.ThreadPoolExecutor() as executor:
        for i in range(1, 5):
            domain = "example%d" % i
            executor.submit(rndc_loop, test_state, domain)

        # Run "rndc status" 10 times, with 1-second pauses between attempts.
        # Each "rndc status" invocation has a timeout of 10 seconds.  If any of
        # them fails, the loop will be interrupted.
        server_is_responsive = True
        attempts = 10
        while server_is_responsive and attempts > 0:
            server_is_responsive = check_if_server_is_responsive()
            attempts -= 1
            time.sleep(1)

        # Signal worker threads that the test is finished.
        test_state["finished"] = True

    # Check whether all "rndc status" commands succeeded.
    assert server_is_responsive



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

from __future__ import print_function
import os
import sys
import signal
import socket
import select
from datetime import datetime, timedelta
import time
import functools

import dns
import dns.edns
import dns.flags
import dns.message
import dns.query
import dns.tsig
import dns.tsigkeyring
import dns.version

from dns.edns import *
from dns.name import *
from dns.rcode import *
from dns.rdataclass import *
from dns.rdatatype import *
from dns.tsig import *


# Log query to file
def logquery(type, qname):
    with open("qlog", "a") as f:
        f.write("%s %s\n", type, qname)


# DNS 2.0 keyring specifies the algorithm
try:
    keyring = dns.tsigkeyring.from_text(
        {
            "foo": {"hmac-sha256", "aaaaaaaaaaaa"},
            "fake": {"hmac-sha256", "aaaaaaaaaaaa"},
        }
    )
except:
    keyring = dns.tsigkeyring.from_text({"foo": "aaaaaaaaaaaa", "fake": "aaaaaaaaaaaa"})

dopass2 = False


############################################################################
#
# This server will serve valid and spoofed answers. A spoofed answer will
# have the address 10.53.0.10 included.
#
# When receiving a query over UDP:
#
# A query to "nocookie"/A will result in a spoofed answer with no cookie set.
# A query to "tcponly"/A will result in a spoofed answer with no cookie set.
# A query to "withtsig"/A will result in two responses, the first is a spoofed
# answer that is TSIG signed, the second is a valid answer with a cookie set.
# A query to anything else will result in a valid answer with a cookie set.
#
# When receiving a query over TCP:
#
# A query to "nocookie"/A will result in a valid answer with no cookie set.
# A query to anything else will result in a valid answer with a cookie set.
#
############################################################################
def create_response(msg, tcp, first, ns10):
    global dopass2
    m = dns.message.from_wire(msg, keyring=keyring)
    qname = m.question[0].name.to_text()
    lqname = qname.lower()
    labels = lqname.split(".")
    rrtype = m.question[0].rdtype
    typename = dns.rdatatype.to_text(rrtype)

    with open("query.log", "a") as f:
        f.write("%s %s\n" % (typename, qname))
        print("%s %s" % (typename, qname), end=" ")

    r = dns.message.make_response(m)
    r.set_rcode(NOERROR)
    if rrtype == A:
        # exempt potential nameserver A records.
        if labels[0] == "ns" and ns10:
            r.answer.append(dns.rrset.from_text(qname, 1, IN, A, "10.53.0.10"))
        else:
            r.answer.append(dns.rrset.from_text(qname, 1, IN, A, "10.53.0.9"))
        if not tcp and labels[0] == "nocookie":
            r.answer.append(dns.rrset.from_text(qname, 1, IN, A, "10.53.0.10"))
        if not tcp and labels[0] == "tcponly":
            r.answer.append(dns.rrset.from_text(qname, 1, IN, A, "10.53.0.10"))
        if first and not tcp and labels[0] == "withtsig":
            r.answer.append(dns.rrset.from_text(qname, 1, IN, A, "10.53.0.10"))
            dopass2 = True
    elif rrtype == NS:
        r.answer.append(dns.rrset.from_text(qname, 1, IN, NS, "."))
    elif rrtype == SOA:
        r.answer.append(dns.rrset.from_text(qname, 1, IN, SOA, ". . 0 0 0 0 0"))
    else:
        r.authority.append(dns.rrset.from_text(qname, 1, IN, SOA, ". . 0 0 0 0 0"))
    # Add a server cookie to the response
    if labels[0] != "nocookie":
        for o in m.options:
            if o.otype == 10:  # Use 10 instead of COOKIE
                if first and labels[0] == "withtsig" and not tcp:
                    r.use_tsig(
                        keyring=keyring,
                        keyname=dns.name.from_text("fake"),
                        algorithm=HMAC_SHA256,
                    )
                elif labels[0] != "tcponly" or tcp:
                    cookie = o
                    if len(o.data) == 8:
                        cookie.data = o.data + o.data
                    else:
                        cookie.data = o.data
                    r.use_edns(options=[cookie])
    r.flags |= dns.flags.AA
    return r


def sigterm(signum, frame):
    print("Shutting down now...")
    os.remove("ans.pid")
    running = False
    sys.exit(0)


############################################################################
# Main
#
# Set up responder and control channel, open the pid file, and start
# the main loop, listening for queries on the query channel or commands
# on the control channel and acting on them.
############################################################################
ip4_addr1 = "10.53.0.9"
ip4_addr2 = "10.53.0.10"
ip6_addr1 = "fd92:7065:b8e:ffff::9"
ip6_addr2 = "fd92:7065:b8e:ffff::10"

try:
    port = int(os.environ["PORT"])
except:
    port = 5300

query4_udp1 = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
query4_udp1.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
query4_udp1.bind((ip4_addr1, port))
query4_tcp1 = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
query4_tcp1.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
query4_tcp1.bind((ip4_addr1, port))
query4_tcp1.listen(1)
query4_tcp1.settimeout(1)

query4_udp2 = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
query4_udp2.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
query4_udp2.bind((ip4_addr2, port))
query4_tcp2 = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
query4_tcp2.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
query4_tcp2.bind((ip4_addr2, port))
query4_tcp2.listen(1)
query4_tcp2.settimeout(1)

havev6 = True
query6_udp1 = None
query6_udp2 = None
query6_tcp1 = None
query6_tcp2 = None
try:
    query6_udp1 = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
    query6_udp1.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    query6_udp1.bind((ip6_addr1, port))
    query6_tcp1 = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
    query6_tcp1.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    query6_tcp1.bind((ip6_addr1, port))
    query6_tcp1.listen(1)
    query6_tcp1.settimeout(1)

    query6_udp2 = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
    query6_udp2.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    query6_udp2.bind((ip6_addr2, port))
    query6_tcp2 = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
    query6_tcp2.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    query6_tcp2.bind((ip6_addr2, port))
    query6_tcp2.listen(1)
    query6_tcp2.settimeout(1)
except:
    if query6_udp1 != None:
        query6_udp1.close()
    if query6_tcp1 != None:
        query6_tcp1.close()
    if query6_udp2 != None:
        query6_udp2.close()
    if query6_tcp2 != None:
        query6_tcp2.close()
    havev6 = False

signal.signal(signal.SIGTERM, sigterm)

f = open("ans.pid", "w")
pid = os.getpid()
print(pid, file=f)
f.close()

running = True

print("Using DNS version %s" % dns.version.version)
print("Listening on %s port %d" % (ip4_addr1, port))
print("Listening on %s port %d" % (ip4_addr2, port))
if havev6:
    print("Listening on %s port %d" % (ip6_addr1, port))
    print("Listening on %s port %d" % (ip6_addr2, port))
print("Ctrl-c to quit")

if havev6:
    input = [
        query4_udp1,
        query6_udp1,
        query4_tcp1,
        query6_tcp1,
        query4_udp2,
        query6_udp2,
        query4_tcp2,
        query6_tcp2,
    ]
else:
    input = [query4_udp1, query4_tcp1, query4_udp2, query4_tcp2]

while running:
    try:
        inputready, outputready, exceptready = select.select(input, [], [])
    except select.error as e:
        break
    except socket.error as e:
        break
    except KeyboardInterrupt:
        break

    for s in inputready:
        ns10 = False
        if s == query4_udp1 or s == query6_udp1 or s == query4_udp2 or s == query6_udp2:
            if s == query4_udp1 or s == query6_udp1:
                print(
                    "UDP Query received on %s"
                    % (ip4_addr1 if s == query4_udp1 else ip6_addr1),
                    end=" ",
                )
            if s == query4_udp2 or s == query6_udp2:
                print(
                    "UDP Query received on %s"
                    % (ip4_addr2 if s == query4_udp2 else ip6_addr2),
                    end=" ",
                )
                ns10 = True
            # Handle incoming queries
            msg = s.recvfrom(65535)
            dopass2 = False
            rsp = create_response(msg[0], False, True, ns10)
            print(dns.rcode.to_text(rsp.rcode()))
            s.sendto(rsp.to_wire(), msg[1])
            if dopass2:
                print("Sending second UDP response without TSIG", end=" ")
                rsp = create_response(msg[0], False, False, ns10)
                s.sendto(rsp.to_wire(), msg[1])
                print(dns.rcode.to_text(rsp.rcode()))

        if s == query4_tcp1 or s == query6_tcp1 or s == query4_tcp2 or s == query6_tcp2:
            try:
                (cs, _) = s.accept()
                if s == query4_tcp1 or s == query6_tcp1:
                    print(
                        "TCP Query received on %s"
                        % (ip4_addr1 if s == query4_tcp1 else ip6_addr1),
                        end=" ",
                    )
                if s == query4_tcp2 or s == query6_tcp2:
                    print(
                        "TCP Query received on %s"
                        % (ip4_addr2 if s == query4_tcp2 else ip6_addr2),
                        end=" ",
                    )
                    ns10 = True
                # get TCP message length
                buf = cs.recv(2)
                length = struct.unpack(">H", buf[:2])[0]
                # grep DNS message
                msg = cs.recv(length)
                rsp = create_response(msg, True, True, ns10)
                print(dns.rcode.to_text(rsp.rcode()))
                wire = rsp.to_wire()
                cs.send(struct.pack(">H", len(wire)))
                cs.send(wire)
                cs.close()
            except s.timeout:
                pass
    if not running:
        break



############################################################################
# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.
############################################################################

# Depends on CWD - Sphinx plugin

import json
from pathlib import Path

import parsegrammar


def read_zone():
    zone_grammars = {}
    for file in Path("../misc/").glob("*.zoneopt"):
        # in-view is not really a zone type
        if file.stem == "in-view":
            zone_type = "in-view"
        else:
            zone_type = f"type {file.stem}"

        with file.open(encoding="ascii") as fp:
            zonegrammar = parsegrammar.parse_mapbody(fp)
            assert len(zonegrammar) == 1
            assert "zone" in zonegrammar
            zone_grammars[zone_type] = zonegrammar["zone"]
            zone_grammars[zone_type]["_pprint_name"] = "zone"

    return {"zone": {"_mapbody": zone_grammars, "_ignore_this_level": True}}


def read_main():
    with Path("../misc/options").open(encoding="ascii") as fp:
        optgrammar = parsegrammar.parse_mapbody(fp)
    return optgrammar


def combine():
    zones = read_zone()
    assert zones
    rest = read_main()
    assert rest
    rest.update(zones)

    # this is a terrible hack
    # but cfg_test cannot print zone grammars inside view
    rest["view"]["_mapbody"].update(zones)

    return rest


if __name__ == "__main__":
    full_grammar = combine()
    print(json.dumps(full_grammar))



#!/usr/bin/python3

# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import os
import re
import subprocess

import pytest

pytest.importorskip("dns", minversion="2.0.0")
import dns.resolver


def run_rndc(server, rndc_command):
    """
    Send the specified 'rndc_command' to 'server' with a timeout of 10 seconds
    """
    rndc = os.getenv("RNDC")
    port = os.getenv("CONTROLPORT")

    cmdline = [rndc, "-c", "../_common/rndc.conf", "-p", port, "-s", server]
    cmdline.extend(rndc_command)

    subprocess.check_output(cmdline, stderr=subprocess.STDOUT, timeout=10)


def test_dnstap_dispatch_socket_addresses(named_port):
    # Prepare for querying ns3.
    resolver = dns.resolver.Resolver()
    resolver.nameservers = ["10.53.0.3"]
    resolver.port = named_port

    # Send some query to ns3 so that it records something in its dnstap file.
    ans = resolver.resolve("mail.example.", "A")
    assert ans[0].address == "10.0.0.2"

    # Before continuing, roll dnstap file to ensure it is flushed to disk.
    run_rndc("10.53.0.3", ["dnstap", "-roll", "1"])

    # Move the dnstap file aside so that it is retained for troubleshooting.
    os.rename(os.path.join("ns3", "dnstap.out.0"), "dnstap.out.resolver_addresses")

    # Read the contents of the dnstap file using dnstap-read.
    output = subprocess.check_output(
        [os.getenv("DNSTAPREAD"), "dnstap.out.resolver_addresses"], encoding="utf-8"
    )

    # Check whether all frames contain the expected addresses.
    #
    # Expected dnstap-read output format:
    #
    #     22-Jun-2022 12:09:06.168 RR 10.53.0.3:0 -> 10.53.0.1:7523 TCP ...
    #     22-Jun-2022 12:09:06.168 RR 10.53.0.3:0 <- 10.53.0.1:7523 TCP ...
    #     22-Jun-2022 12:09:06.168 RQ 10.53.0.3:56306 -> 10.53.0.2:7523 UDP ...
    #     22-Jun-2022 12:09:06.168 RQ 10.53.0.3:56306 <- 10.53.0.2:7523 UDP ...
    #
    bad_frames = []
    inspected_frames = 0
    addr_regex = r"^10\.53\.0\.[0-9]+:[0-9]{1,5}$"
    for line in output.splitlines():
        _, _, frame_type, addr1, _, addr2, _ = line.split(" ", 6)
        # Only inspect RESOLVER_QUERY and RESOLVER_RESPONSE frames.
        if frame_type not in ("RQ", "RR"):
            continue
        inspected_frames += 1
        if not re.match(addr_regex, addr1) or not re.match(addr_regex, addr2):
            bad_frames.append(line)

    assert (
        len(bad_frames) == 0
    ), "{} out of {} inspected frames contain unexpected addresses:\n\n{}".format(
        len(bad_frames), inspected_frames, "\n".join(bad_frames)
    )



# Copyright (C) Internet Systems Consortium, Inc. ("ISC")
#
# SPDX-License-Identifier: MPL-2.0
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0.  If a copy of the MPL was not distributed with this
# file, you can obtain one at https://mozilla.org/MPL/2.0/.
#
# See the COPYRIGHT file distributed with this work for additional
# information regarding copyright ownership.

import pytest

pytest.importorskip("dns")
import dns.message
import dns.query


@pytest.mark.parametrize(
    "qname,rdtype,expected_ttl",
    [
        ("min-example.", "SOA", 60),
        ("min-example.", "MX", 30),
        ("max-example.", "SOA", 120),
        ("max-example.", "MX", 60),
    ],
)
def test_cache_ttl(qname, rdtype, expected_ttl, named_port):
    msg = dns.message.make_query(qname, rdtype)
    response = dns.query.udp(msg, "10.53.0.2", timeout=10, port=named_port)
    for rr in response.answer + response.authority:
        assert rr.ttl == expected_ttl
